
import warnings
import numpy as np
from scipy import stats
import pandas
from . import io
from astropy import time, cosmology
_COEFS = 1.57
_COSMO = cosmology.Planck18
# Top level method #

def get_sample(**kwargs):
    """ Short to to Sample.load() """
    return Sample.load(**kwargs)


def get_data( saltmodel="default",
              good_coverage = True,
              good_lcfit = True,
              redshift_source=None,
              redshift_range=None,
              get_envprop = True, corr_mw=True,
                 **kwargs):
    """ Generic dataframe for the ZTF Cosmo DR2 sample.
   
    - good_coverage=True, means minimal sampling (see overview paper)
    - good_lcfit=True, means:
       - x1_range=[-3, 3],
       - c_range=[-0.2, 0.8],
       - t0_err_range=[0, 1],
       - x1_err_range=[0, 1], 
       - c_err_range=[0, 0.1],
       - fitprob > 1e-7

    use **kwargs to change any selection (see sample.get_data())
    """
    from .hostmass import get_taylor_hostmass
    sample = get_sample(saltmodel=saltmodel)
    
    prop = { **dict(good_coverage=good_coverage,
                    good_lcfit=good_lcfit,
                    redshift_range=redshift_range,
                    redshift_source =redshift_source),
              **kwargs}
    
    data = sample.get_data(**prop)
    
    # Hubble Residuals
    data['mb'] = -2.5 * np.log10( data["x0"] ) + 19*_COEFS
    data["mb_err"] =  +2.5/np.log(10) * data["x0_err"] / data["x0"]
    
    data['magres'] = data['mb'] - _COSMO.distmod(data["redshift"].values).value
    data["magres_err"] =  data["mb_err"].copy()
    data["cov_mag_c"] = -2.5*np.array(data['cov_x0_c'])/(np.log(10)*data['x0'])
    data["cov_mag_x1"] = -2.5*np.array(data['cov_x0_x1'])/(np.log(10)*data['x0'])
    
    # Environmental data
    if get_envprop:
        localmags = io.get_localhost_mag()
        globalmags = io.get_globalhost_mag().join(data[["redshift"]])

        if corr_mw:
            from .mw import correct_mwextinction_on_hostmag
            localmags = localmags.join(data["mwebv"])
            globalmags = globalmags.join(data["mwebv"])
            # apply the correction
            localmags = correct_mwextinction_on_hostmag(localmags)
            globalmags = correct_mwextinction_on_hostmag(globalmags)
            
        # get the host masses using Taylor's relation
        host_info = globalmags[["PS1i","PS1g","PS1i_err","PS1g_err", "redshift"]].dropna()
        mass, mass_err = get_taylor_hostmass(magi = host_info["PS1i"], magg = host_info["PS1g"],
                                         magi_err = host_info["PS1i_err"], magg_err = host_info["PS1g_err"],
                                         redshift=host_info["redshift"], use_prior=True)
        masses = mass.to_frame("mass").join(mass_err.to_frame("mass_err")).reindex(data.index)

        # get the color colors
        localmags["localcolor"] = localmags["PS1g"] - localmags["PS1z"]
        localmags["localcolor_err"] = np.sqrt(localmags["PS1g_err"]**2 + localmags["PS1z_err"]**2)

    
        #localmags["mass"] = localmags["PS1g"] - localmags["PS1z"]
        data = data.join(localmags[["localcolor","localcolor_err"]].loc[data.index])
        data = data.join(masses[["mass","mass_err"]].loc[data.index])
        data = data.join(globalmags[["PS1r"]].loc[data.index])
        
        ## Standardisation parameters
        lcolor_cut = 1
        data["h_lowcolor"] = stats.norm.cdf(lcolor_cut, loc=data["localcolor"], scale=data["localcolor_err"])
        data["h_lowcolor_err"] = 1e-3
    
        data["h_lowmass"] = stats.norm.cdf(10, loc=data["mass"], scale=data["mass_err"])
        data["h_lowmass_err"] = 1e-3

    return data.copy()


def _get_coverage_(phase_df, prefix=None, one_per_day=False):
    """ """
    if prefix is None:
        prefix = ""

    if one_per_day:
        phase_df = phase_df.copy().astype(int).reset_index().drop_duplicates().set_index(["ztfname","filter"])["phase"]
        
    n_points  = phase_df.groupby(level=0).size().to_frame(f"n_{prefix}points")
    n_bands   = ((phase_df.groupby(level=[0,1]).size()>0).groupby(level=0).sum()).to_frame(f"n_{prefix}bands")
    n_points_g = phase_df.xs("ztfg", level=1).groupby(level=0).size().to_frame(f"n_{prefix}points_ztfg")
    n_points_r = phase_df.xs("ztfr", level=1).groupby(level=0).size().to_frame(f"n_{prefix}points_ztfr")
    n_points_i = phase_df.xs("ztfi", level=1).groupby(level=0).size().to_frame(f"n_{prefix}points_ztfi")
    dfs = [n_points, n_bands, n_points_g, n_points_r, n_points_i]
    return pandas.concat(dfs, axis=1).fillna(0).astype( int )

class Sample():

    
    def __init__(self, data=None, saltmodel=None):
        """ """
        self.set_data(data)
        self._saltmodel = saltmodel
        
    @classmethod
    def load(cls, redshift_range=None, target_list=None, has_spectra=True,
                 saltmodel="default", **kwargs):
        """ Load a Sample instance building it from io.get_targets_data() """
        data, saltmodel = io.get_targets_data(saltmodel=saltmodel, **kwargs)
        
        if redshift_range is not None:
            data = data[data["redshift"].between(*redshift_range)]

        if target_list is not None:
            data = data.loc[target_list]

        if has_spectra:
            specfile = io.get_spectra_datafile(data=data)
            data = data[data.index.isin(specfile["ztfname"])]
            
        return cls(data=data, saltmodel=saltmodel)

    # ------- #
    # LOADER  #
    # ------- #
    def load_hostdata(self):
        """ load the host data using io.get_host_data(). This is made automatically upon hostdata call. """
        self.set_hostdata( io.get_host_data() )

    def load_spectra_df(self, add_phase=True, current_target=True):
        """ load the spectra dataframe 

        Parameters
        ----------
        add_phase: bool
            should the phase (based on spectra date and the target t0) be added

        current_target: bool
            only keep target that are in the current sample.
        
        Returns
        -------
        None
            sets self.spectra_df
        """
        spectra_df = io.get_spectra_datafile().set_index("ztfname")
        if current_target:
            spectra_df = spectra_df.loc[spectra_df.index.isin( self.data.index )]
            
        spectra_df["mjd"] = time.Time(np.asarray(spectra_df["date"].apply(lambda x: f"{x[:4]}-{x[4:6]}-{x[6:]}").astype(str).values, dtype="str")).mjd
        if add_phase:
            spectra_df = spectra_df.join( self.data["t0"] )
            spectra_df["phase"] = spectra_df["mjd"] - spectra_df["t0"]

        self._spectra_df = spectra_df
        
    def load_phase_df(self, min_detection=5, groupby='filter', client=None, rebuild=False, **kwargs):
        """ Load the phase_df. This is made automatically upton phase_df call.
        If this is the first time ever you call this, the phase_df has to be built, so it takes ~45s.
        Once built, the dataframe is stored such that, next time, it is directly loaded. 
        Use rebuild=True to bypass the automatic loading and rebuild the dataframe.
        
        Parameters
        ----------
        min_detection: [float] -optional-
            minimal signal to noise ratio for a point to be considered as 'detected'

        groupby: [string or list of] -optional-
            data column(s) to group the lightcurve data and measuring the count statistics.
            example: filter, [filter, fieldid]
             
        client: [dask.Client] -optional-
            if a dask.distributed.Client instance is given, this will use dask. 
            Otherwise a basic for loop (slower) is used.

        rebuild: [bool] -optional-
            force the dataframe to be rebuilt.

        **kwargs goes to self.build_phase_coverage()

        Returns
        -------
        None
        """
        if not rebuild:
            phase_df = io.get_phase_coverage(load=True, warn=False)
            if phase_df is None:
                rebuild = True
                
        if rebuild:
            phase_df = self.build_phase_coverage(min_detection=min_detection,
                                                    groupby=groupby,
                                                    client=client, store=True, **kwargs)
        self._phase_df = phase_df
        
    def load_fiedid(self):
        """ compute the fields containing the targets using 
        ztfquery.fields.get_fields_containing_target().
        This take quite some time.

        the 'fieldid' entry is added to self.data
        """
        from ztfquery import fields
        fieldid = [fields.get_fields_containing_target(s_["ra"],s_["dec"])
                       for i_,s_ in self.data.iterrows()]
        self.data["fieldid"] = fieldid

    # ------- #
    # SETTER  #
    # ------- #
    def set_data(self, data):
        """ attach to this instance the target data. 
        = Most likely you should not use this method directly. =
        use sample = Sample.load()
        """
        try:
            data["t0day"] = data["t0"].astype("int")
        except:
            data["t0day"] = data["t0"]
            
        self._data = data

    def set_hostdata(self, hostdata):
        """ attach to this instance the host data.
        = Most likely you should not use this method directly. =
        use sample.load_hostdata() or simply 
        sample.hostdata that automatically load this corrent hostdata.
        """
        self._hostdata = hostdata
        
    def merge_to_data(self, dataframe, how="outer", **kwargs):
        """ Merge the given dataframe with self.data. 
        The merged dataframe will replace self.data
        """
        self._data = pandas.merge(self.data, dataframe, how=how, 
                                 **{**dict(left_index=True, right_index=True),
                                 **kwargs})
        
    # ------- #
    # GETTER  #
    # ------- #
    def get_data(self,
                     good_coverage=None,
                     good_lcfit=None,
                     clean_t0nan=True,                     
                     t0_range=None, x1_range=None, c_range=None,
                     redshift_range=None,
                     redshift_source=None,
                     t0_err_range=None, c_err_range=None, x1_err_range=None,
                     exclude_targets=None, in_targetlist=None,
                     ndetections=None,
                     fitprob=None,
                     classification=None,
                     first_spec_phase=None, query=None, data=None):
        """ 
        *_range: [None or [min, max]] -optional-
            cut to be applied to the data for
            t0, x1, c, (and there error) and redshift.
            boundaries are mandatory. For instance, redshift range lower than 0.06
            should be: redshift_range = (0, 0.06).
            = see below for t0 format =
        
        t0_range: [None or [tmin,tmax]] -optional-
            Should be a format automatically understood by astropy.time.Time
            e.g. t0_range=["2018-04-01","2020-10-01"]

        in_targetlist: [list of strings] -optional-
            The target must be in this list

        goodcoverage: [None or bool] -optional-
            Select the data given the lc phase coverage
            - None: no cut
            - True: only good lc kept
            - False: only bad lc kept (good lc discarded)
            This uses self.get_goodcoverage_targets(**coverage_prop)

        coverage_prop: [dict] -optional-
            kwargs passed to self.get_goodcoverage_targets
            = used only if goodcoverage is not None =

        query: [string] -optional-
            any additional query to be given to data.query({query}).
            This are SQL-like format applied to the colums. 
            See pandas.DataFrame.query()


        Returns
        -------
        DataFrame (sub part or copy of self.data)
        """
            
        if clean_t0nan:
            data = self.data[~self.data["t0"].isna()]
        else:
            data = self.data.copy()
        
        # - Time Range
        if t0_range is not None:
            t0_start = time.Time(t0_range[0]).mjd
            t0_end = time.Time(t0_range[1]).mjd
            data = data[data["t0"].between(t0_start, t0_end)]

        # LC Cuts
        # - stretch (x1) range
        if x1_range is not None:
            data = data[data["x1"].between(*x1_range)]

        # - color (c) range
        if c_range is not None:
            data = data[data["c"].between(*c_range)]

        # - t0 errors  range
        if t0_err_range is not None:
            data = data[data["t0_err"].between(*t0_err_range)]
            
        # - stretch errors (x1) range
        if x1_err_range is not None:
            data = data[data["x1_err"].between(*x1_err_range)]

        # - color errors (c) range
        if c_err_range is not None:
            data = data[data["c_err"].between(*c_err_range)]


        # Redshift Cuts
        # - Redshift range
        if redshift_range is not None:
            data = data[data["redshift"].between(*redshift_range)]
            
        # -  redshift origin
        if redshift_source is not None:
            data = data[data["source"].isin(np.atleast_1d(redshift_source))]


        # Classification
        if classification is not None:
            data = data[data["sn_type"].isin(np.atleast_1d(classification))]
            
        # Target cuts
        # - in given list
        if in_targetlist is not None:
            data = data.loc[np.asarray(in_targetlist)[np.in1d(in_targetlist, data.index.astype("string"))] ]

        if exclude_targets is not None:
            data = data.loc[~data.index.isin(exclude_targets)]


        # LC Coverage
        if ndetections is not None:
            phase_coverage = self.get_phase_coverage()["n_points"]
            min_corevarge = phase_coverage[phase_coverage>=ndetections]
            data = data.loc[data.index.isin(min_corevarge.index)]
            
        # - special good lc list.            
        if good_coverage is not None:
            if good_coverage:
                data = data[data["lccoverage_flag"].astype(bool)]
            else:
                data = data[~data["lccoverage_flag"].astype(bool)]


        if good_lcfit is not None:
            if good_lcfit:
                data = data[data["fitquality_flag"].astype(bool)]
            else: # lcquality_flag
                data = data[~data["fitquality_flag"].astype(bool)]
                
        if fitprob is not None:
            data = data[data["fitprob"]>fitprob]
                
        # Spectral Cut.
        if first_spec_phase is not None:
            first_spec = self.spectra_df.groupby(level=0).phase.first()
            first_spec = first_spec[first_spec<=first_spec_phase]
            data = data.loc[data.index.isin(first_spec.index)]

        # Additional Query
        if query:
            data = data.query(query)


        bitmask = np.zeros(len(data))
        for i, (value, range) in enumerate( [["x1", [-3,3]], #2**0
                                     ["c", [-0.2, 0.8]],#2**1
                                     ["t0_err", [0, 1]],#2**2
                                     ["x1_err", [0, 1]],#2**3
                                     ["c_err", [0, 0.1]],#2**4
                                     ["fitprob", [1e-7, 10]]]):#2**5
            flag_in = data[value].between(*range).values
            bitmask[~flag_in] += 2**i
        data["fitquality_bitmask"] = np.asarray(bitmask, dtype=int)
    
        return data

    def get_phases(self, one_per_day=False,
                         phase_range=None,
                         detection=5, incl_errorfloor=True,
                         in_targetlist=None,
                         flagout=[1, 2, 4, 8, 16],
                         restframe=True):
        """ """
        phase_df = self.phase_df.copy()
        
        if restframe:
            phase_df = phase_df.join(self.data["redshift"])
            phase_df["phase"] = phase_df["phase"]/(1+phase_df["redshift"])

        if phase_range is not None:
            phase_df = phase_df[phase_df["phase"].between(*phase_range)]
        
        # remove bad lc points
        if flagout is not None:
            flag_ = np.all([(phase_df["flag"]&i_==0) for i_ in np.atleast_1d(flagout)], axis=0)
            phase_df = phase_df[flag_]


        if detection is not None:
            if incl_errorfloor:
                det_key = "detection_ef"
            else:
                det_key = "detection"
            phase_df = phase_df[(phase_df[det_key]>=detection)]["phase"].copy()
        else:
            phase_df = phase_df["phase"]

        if one_per_day:
            phase_df = phase_df.astype(int).reset_index().drop_duplicates().set_index(["ztfname","filter"])["phase"]
            
        if in_targetlist:
            phase_df = phase_df[phase_df.index.get_level_values(0).isin(np.atleast_1d(in_targetlist))]
            
        return phase_df
                                    
    
    # Target | High level
    def get_target(self, name):
        """ """
        from . import target
        lightcurve = self.get_target_lightcurve(name)
        spectra = self.get_target_spectra(name, as_spectra=False)
        meta = self.data.loc[name]
        return target.Target(lightcurve, spectra, meta=meta)

    def get_target_typing(self, name=None):
        """ """
        if name is None:
            return self.data["sn_type"].copy()
        
        return self.data.loc[np.atleast_1d(name)]["sn_type"].copy()

    # model
    def get_target_saltmodel(self, name):
        """ """
        from .lightcurve import get_saltmodel
        propmodel = self.data.loc[name].rename({"redshift":"z"})[["z","t0","x0","x1","c","mwebv"]].to_dict()
        return get_saltmodel(which=self._saltmodel, **propmodel)
        
    # LightCurve
    def get_target_lightcurve(self, name, **kwargs):
        """ Get the {name} LightCurve object """
        from . import lightcurve
        return lightcurve.LightCurve.from_name(name, saltparam=self.data.loc[name], saltmodel=self._saltmodel)

    def get_target_lightcurve_data(self, name,  phase_range=None, mjd_range=None, **kwargs):
        """ Get the {name} LightCurve object """
        from . import lightcurve
        return lightcurve.get_target_lcdata(name, saltparam=self.data.loc[name],  phase_range=None, mjd_range=None, **kwargs)
    
    def get_target_lightcurve_residual(self, name, phase_range=None, restphase=False, mjd_range=None, **kwargs):
        """ fet the {name} lightcurve residuals given the target's salt model. 
        
        Parameters
        ----------
        name: str
            target's name

        phase_range: [float, float]
            limit phase range (days to maximum of light t0): [start, end]
        
        **kwargs goes to lightcurve.LightCurve.get_model_residual

        Returns
        -------
        pandas.DataFrame
        """
        from . import lightcurve
        if phase_range is not None and restphase:
            phase_range = np.asarray(phase_range)/(1+self.data.loc[name]["redshift"])
        
        return lightcurve.get_target_lcresiduals(name, phase_range,
                                                     mjd_range=mjd_range,
                                                     saltparam=self.data.loc[name],
                                                     which_model=self._saltmodel,
                                                     **kwargs)
    
    # Spectrum
    def get_target_spectra(self, name, **kwargs):
        """ Get a list with all the Spectra for the given object """
        from . import spectroscopy
        return spectroscopy.Spectrum.from_name(name, **kwargs)

    # Extra
    def get_goodcoverage_targets(self, detection=5, one_per_day=True,
                                      **kwargs):
        """ kwargs should have the same format as the n_early_point='>=2' for instance.
        None means no constrain, like n_bands=None means 'n_bands' is not considered.
        """
        # Only considering 1 point per band and per day (mjd int) if one_per_day==True
        query = {**dict(n_points=">=7", # any band
                        n_early_points=">=2", # e.g. at least 2g or 1g and 1i
                        n_late_points=">=2", # e.g. at least 2g or 1g and 1i
                        n_bands=">=2", # e.g. at least 1g and 1r | 2 filters post max
                        ),
                 **kwargs}
        df_query = " and ".join([f"{k}{v}" for k,v in query.items() if v is not None])

        phase_coverage = self.get_phase_coverage(detection=detection,
                                                 one_per_day=one_per_day)
        
        return phase_coverage.query(df_query).index.astype("string")
    
    def get_phase_coverage(self, premax_range = [-10, 0],
                                 postmax_range = [0, 40],
                                 phase_range = [-10, 40],
                                 one_per_day=True,
                                 detection=5):
        """ """
        # force phase range first then bin per day | don't been yet
        phase_df = self.get_phases(one_per_day=False,
                                   detection=detection)

        dfs = _get_coverage_( phase_df[phase_df.between(*phase_range)],
                                  one_per_day=one_per_day
                            )
        
        dfs_early = _get_coverage_(phase_df[phase_df.between(*premax_range)],
                                       prefix="early_", one_per_day=one_per_day)
        dfs_late = _get_coverage_(phase_df[phase_df.between(*postmax_range)],
                                      prefix="late_", one_per_day=one_per_day)
        
        return pandas.concat([dfs, dfs_early, dfs_late], axis=1
                            ).reindex(self.data.index).fillna(0).astype(int)
        
    def build_phase_coverage(self, groupby='filter', client=None, store=True,
                                 error_floor={"ztfg":2.5, "ztfr":3.5, "ztfi":6},
                                 **kwargs):
        """ 
        time: 
        - Dask: 45s on a normal laptop | 4 cores 
        - No Dask: 60s on a normal laptop | 4 cores 
        """
        warnings.warn("building phase coverage takes ~30s.")
        
        import pandas
        from . import io
        phases = []
        data = self.get_data()
        
        lcdata = io.get_lightcurve_datafile(contains=".csv").set_index("ztfname").loc[data.index]
        dfs = pandas.concat([pandas.read_csv(f_, delim_whitespace=True, comment='#') for f_ in lcdata["fullpath"]],
                                keys=data.index)
        dfs["phase"] = dfs["mjd"] - data["t0"].reindex(dfs.index, level=0)
        dfs["detection"] = dfs["flux"]/dfs["flux_err"]
        # error floor:
        error_floor=pandas.Series(error_floor, name="errorfloor").to_frame().reset_index().rename({"index":"filter"},
                                                                                                   axis=1)
        dfs = dfs.reset_index().merge(error_floor, on="filter")
        dfs["detection_ef"] = dfs["flux"]/ (np.sqrt(dfs["flux_err"]**2 + (dfs["flux"]*dfs["errorfloor"]/100)**2) )
        phase_df = dfs.reset_index().set_index(["ztfname","filter"])[["phase","detection",
                                                                          "detection_ef",
                                                                      "flag","in_baseline"]]
        if store:
            filepath = io.get_phase_coverage(load=False)
            phase_df.to_parquet(filepath)
            
        return phase_df
    
    
    # ------- #
    # PLOTTER #
    # ------- #
    def show_ideogram(self, key, keyerr=None,
                          ax=None, data=None, dataprop={},
                          facecolor=None, edgecolor=None, lw=2,
                          nsigma_range=3, npoints=1000,
                          density=False, normed_one=False,
                          **kwargs):
        """ """
        from matplotlib.colors import to_rgba

        # axes
        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[6,3])
            ax = fig.add_axes([0.1,0.2,0.8,0.7])
        else:
            fig = ax.figure

        # data            
        if keyerr is None:
            keyerr = key+"_err"

        if data is None:
            data = self.get_data(**dataprop)

        # getting the x and y            
        d, d_err = data[[key, keyerr]].values.T
        d_range = np.nanmin(d-nsigma_range*d_err), np.nanmax(d+nsigma_range*d_err)
        dd = np.linspace(*d_range, npoints)
        dideo = np.sum(stats.norm.pdf(dd[:,None],
                                      loc=d, scale=d_err), 
                           axis=1)
        
        # plotting properties
        if facecolor is None:
            facecolor = to_rgba("C0",0.3)
            facecolor_ = "C0"
        else:
            facecolor_ = facecolor
            
        if edgecolor is None:
            edgecolor = to_rgba(facecolor_, 1)


        if density:
            dideo /= np.nansum(dideo)
            
        if normed_one:
            dideo /= np.nanmax(dideo)
            
        # plot
        _ = ax.fill_between(dd, dideo, 
                                facecolor=facecolor, 
                                edgecolor=edgecolor, lw=lw,
                                **kwargs)
        
        # fancy
        ax.set_ylim(0)
        return fig
    
    def show_discoveryhist(self, ax=None, daymax=15, linecolor="C1", **kwargs):
        """ """
        from matplotlib.colors import to_rgba
        datasalt = self.get_data(clean_t0nan=True)
        
        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[6,3])
            ax = fig.add_axes([0.1,0.2,0.8,0.7])
        else:
            fig = ax.figure
            

        prop = dict(fill=True, histtype="bar", 
                    density=False, facecolor=to_rgba("C0", 0.1), edgecolor="C0",
                   align="left", zorder=3)
        _ = ax.hist(datasalt.groupby("t0day").size(), range=[0,15], bins=15,
                    **{**prop,**kwargs})
        if linecolor != "None":
            ax.axvline(3.4, color="C1", zorder=5, lw=1.5, ls="--")
        
#        xx = np.arange(0, daymax)        
        #ax.scatter(xx[1:], stats.poisson.pmf(xx[1:], mu=3.5)*1050, 
        #           s=50, color="C1", zorder=5)

        #ax.set_yscale("log")

        _ = ax.set_xticks(np.arange(daymax))
        ax.set_xlim(0)

        clearwhich = ["left","right","top"] # "bottom"
        [ax.spines[which].set_visible(False) for which in clearwhich]
        ax.tick_params("y", labelsize="x-small")
        ax.set_yticks([0,100,200])

        ax.set_xlabel("Number of SN Ia per day")
        return fig
    
    
    def show_discoveryevol(self, ax=None, t0_range=["2018-04-01","2021-01-01"],
                           typed_color="C0", quality_color="goldenrod",
                           xformat=True, dataprop={}, 
                           **kwargs):
        """ """
        from matplotlib.colors import to_rgba

        # - Data    
        datasalt_all = self.get_data(clean_t0nan=True, 
                                        t0_range=t0_range,
                                        **dataprop)
        
        datasalt_lccut = self.get_data(clean_t0nan=True,
                                        t0_range=t0_range,
                                        goodcoverage=True,
                                        **dataprop)
        
        datasalt_zcut = self.get_data(clean_t0nan=True,
                                        t0_range=t0_range, 
                                        z_quality=2,
                                        **dataprop)
        
        datasalt_zcut_lccut = self.get_data(clean_t0nan=True,
                                        t0_range=t0_range, 
                                        z_quality=2,
                                        goodcoverage=True,
                                        **dataprop)

        
        # - Figue
        if ax is None:
            import matplotlib.pyplot as mpl            
            fig = mpl.figure(figsize=[8,4])
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure


        # - Internal

        def _show_single_(gb_cumsize, mplfunc="fill_between", 
                          add_text=True, text_color=None, **kwargs_):
            """ """
            time_ = time.Time(gb_cumsize.index, format="mjd").datetime
            values_ = gb_cumsize.values
            _ = getattr(ax,mplfunc)(time_, values_, **kwargs_)
            if add_text:
                ax.text(time_[-1], values_[-1],f" {values_[-1]}",
                            va="center",ha="left", color=text_color)

        # Show all
        _show_single_(datasalt_all.groupby("t0day").size().cumsum(), 
                     lw=2, ls="None", facecolor="w", edgecolor="None", 
                      add_text=False)
        _show_single_(datasalt_all.groupby("t0day").size().cumsum(), 
                     lw=2, ls="--", color=typed_color, mplfunc="plot",
                     text_color=typed_color)



        _show_single_(datasalt_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="None", facecolor=to_rgba(quality_color,0.2), 
                      edgecolor="None", add_text=False)
        _show_single_(datasalt_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="--", color=quality_color, mplfunc="plot",
                     text_color=quality_color)



        _show_single_(datasalt_zcut.groupby("t0day").size().cumsum(), 
                      mplfunc="plot",
                     lw=1, ls="-", color=typed_color, 
                     text_color=typed_color)

        _show_single_(datasalt_zcut_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="-", facecolor=to_rgba(quality_color,0.2), 
                      edgecolor="None", add_text=False)
        _show_single_(datasalt_zcut_lccut.groupby("t0day").size().cumsum(), 
                     lw=2, ls="-", color=quality_color, mplfunc="plot",
                     text_color=quality_color)


        # - Out Formating
        if xformat:
            from matplotlib import dates as mdates
            locator = mdates.AutoDateLocator()
            formatter = mdates.ConciseDateFormatter(locator)
            ax.xaxis.set_major_locator(locator)
            ax.xaxis.set_major_formatter(formatter)

        clearwhich = ["left","right","top"] # "bottom"
        [ax.spines[which].set_visible(False) for which in clearwhich]

    #        ax.set_title("ZTF-1 Ia Statistics")
        ax.set_ylabel("Number of Type Ia Supernovae")
        ax.set_ylim(bottom=0)
        ax.tick_params("y", labelsize="small")
        return fig



        
    # =============== #
    #   Properties    #
    # =============== #
    @property
    def data(self):
        """ """
        return self._data

    @property
    def hostdata(self):
        """ """
        if not hasattr(self, "_hostdata"):
            self.load_hostdata()
        return self._hostdata
    
    @property
    def phase_df(self):
        """ """
        if not hasattr(self, "_phase_df"):
            self.load_phase_df()
        
        return self._phase_df

    @property
    def phase_coverage(self):
        """ """
        if not hasattr(self, "_phasecoverage"):
            self._phasecoverage = self.get_phase_coverage() # default
            
        return self._phasecoverage
    
    @property
    def spectra_df(self):
        """ dataframe containing the spectral information """
        if not hasattr(self, "_spectra_df"):
            self.load_spectra_df()
        
        return self._spectra_df




# Loads the ztf filters inside sncosmo
import warnings
from ztfquery import filters
import numpy as np
import pandas
filters.load_p48_filters_to_sncosmo(basename="p48") # p48g etc.


from .io import get_saltparams
try:
    SALT2PARAMS = get_saltparams()
except:
    warnings.warn("Failed to load SALT2 parameters")
    SALT2PARAMS = None



def salt2result_to_dataframe(result):
    """ """
    fitted = np.in1d(result.param_names, result.vparam_names)
    df = pandas.DataFrame(np.asarray([result.parameters, fitted]).T, 
                          result.param_names, 
                          columns=["values", "fitted"])
    df.fitted = df.fitted.astype(bool)
    
    # - Error
    df = df.merge(pandas.Series(dict(result["errors"]), name="errors"), 
                      left_index=True, right_index=True, how="outer")
    
    # - Cov
    dcov = pandas.DataFrame(result["covariance"], columns=result.vparam_names, index=result.vparam_names)
    dcov.columns ="cov_"+dcov.columns
    
    # - merged
    return df.merge(dcov,  left_index=True, right_index=True, how="outer")
    




import os
import warnings
import pandas
import numpy as np
IDR_PATH = os.getenv("ZTFIDRPATH", "./dr2")

__all__ = ["get_targets_data"]
    
# ================== #
#                    #
#   TOP LEVEL        #
#                    #
# ================== #
def get_targets_data(saltmodel="default"):
    """ 
    Returns
    -------
    pandas.DataFrame, string
        - dataframe containing targets parameters (incl lc fit parameters)
        - name of the lc fit model.
    """
    redshifts = get_redshif_data()[["redshift","redshift_err", "source"]]
    saltparams, saltmodel = get_saltparams(which=saltmodel)
    coords = get_coords_data()
    
    # merging
    data_ = pandas.merge(redshifts,saltparams, left_index=True, right_index=True,
                     suffixes=("","_salt"), how="outer")
    data_ = pandas.merge(data_, coords, left_index=True, right_index=True,
                         how="outer")
    
    # force limit to target to use
    target_list = get_targetlist()
    data_ = data_.loc[target_list["ztfname"]]

    # adding classification
    typing = get_target_typing()
    data_ = data_.join(typing[["sn_type", "sub_type"]], how="left")
    return data_, saltmodel

# ================== #
#                    #
#   BASICS           #
#                    #
# ================== #
def get_targetlist(load=True, **kwargs):
    """ official list of target to use for dr2 """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_targetlist.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, **kwargs)

def get_target_typing(load=True, index_col=0, sep=" ",
                        from_datacreation=False,
                        clean=True, generic_typing=True,
                          **kwargs):
    """ """
    if not from_datacreation: # bases:
        filepath = os.path.join(IDR_PATH, "tables/ztfdr2_classifications.csv")
        sep = ","
    else:
        filepath = os.path.join(IDR_PATH, "tables/.dataset_creation/sample_def",
                            "ztfdr2_typing.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, index_col=index_col, sep=sep, **kwargs)
    if not from_datacreation:
        return data
    
    if clean:
        data["typing"] = data["typing"].str.replace("[","", regex=False).str.replace("]","", regex=False).str.replace("'","") # clean
        data["typing"] = data["typing"].str.replace("sn ia", "snia").str.replace("not ia", "notia")
        data["typing"] = data["typing"].str.split(" ")
        
        data["ntypings"] = data["ntypings"].str.replace("[","", regex=False).str.replace("]","", regex=False).str.split(" ")

    if generic_typing:
        TYPING = {'sn ia': 'snia',
                  'ia-norm': 'snia-norm',
                  'ia(-norm)': 'snia-norm',
                  'ia-91t':'snia-pec-91t',
                  'ia-91bg':'snia-pec-91bg',
                  'ia-other':'snia-pec'}

        data["classification"] = data["classification"].apply(TYPING.get)
        
    return data

# Master List
def get_masterlist(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/sample_def/ztfdr2_masterlist.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, **kwargs)
    return data["ztfname"].values.astype(str)

def get_nospectralist(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/sample_def/ztfdr2_target_nospectra.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, names=["ztfname"], **kwargs)
    return data["ztfname"].values.astype(str)


# Redshifts
def get_redshif_data(load=True, index_col=0, full=False, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_redshifts.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, index_col=index_col, **kwargs)
    data.index.name = 'ztfname'

    if not full: # get only redshift and source
        return data
    
    # or get them all:
    redshift_dir = os.path.join(IDR_PATH, "tables", ".dataset_creation/redshifts")
    redshifts = {}

    
    # Emission lines
    sedm_em = pandas.read_csv(os.path.join(redshift_dir,
                                            "ztfdr2_sedmhanii_redshift.csv"),
                                  index_col=0)
    sedm_em.columns = sedm_em.columns.str.replace("redshift","sedm")
    
    
    
    nosedm_em = pandas.read_csv(os.path.join(redshift_dir,
                                                    "ztfdr2_nonsedmhanii_redshift.csv"),
                                     index_col=0)
    nosedm_em.columns = nosedm_em.columns.str.replace("redshift","nonsedm")
    
    # SNID
    snidauto = pandas.read_csv(os.path.join(redshift_dir,
                                                    "ztfdr2_snid_redshifts.csv"),
                                     index_col=0)
    snidauto.columns = snidauto.columns.str.replace("redshift", "snid")
    
    # Host-Z
    hostz = pandas.read_csv(os.path.join(redshift_dir,
                                                    "ztfdr2_hostphot_redshifts.csv"),
                                     index_col=0)
    hostz = hostz[["z"]]
    hostz.columns = ["gal"]
    hostz["gal_err"] = 1e-5


    
    # override
#    override = pandas.read_csv(os.path.join(redshift_dir,
#                                                    "ztfdr2_override_redshifts.csv"),
#                                     index_col=0)
#    override.columns = ["override"]

    # merge them
    
    data = data.join(sedm_em, on="ztfname")
    data = data.join(nosedm_em, on="ztfname")
    data = data.join(snidauto, on="ztfname")
    data = data.join(hostz, on="ztfname")
 #   data = data.join(override, on="ztfname")
    data
    
    return data
    
# Coordinates
def get_coords_data(load=True, index_col=0, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_coordinates.csv")
    if not load:
        return filepath
    return pandas.read_csv(filepath, index_col=index_col, **kwargs)

# SALT
def get_saltparams(load=True, which="default", **kwargs):
    """ 

    Parameters
    ----------
    which: string
        name of the salt2 fit file to load. 
        If 'default': ztfdr2_salt2_params.csv is used
        
    Returns
    -------
    pandas.DataFrame, string
        - dataframe containing targets lc fit parameters
        - name of the lc fit model.
    """
    if which == "default":
        filename = "ztfdr2_salt2_params.csv"
    else:
        filename = os.path.join(".dataset_creation/lc_fits", which)

    filepath = os.path.join(IDR_PATH, "tables", filename)
    saltmodel = _parse_salt2filename(filepath)
    
    
    if not load:
        return filepath, saltmodel
    
    return pandas.read_csv(filepath, **kwargs
                          ).rename({"z":"redshift"}, axis=1
                          ).set_index("ztfname"), saltmodel


def _parse_salt2filename(filename):
    """ """
    basename = os.path.basename(filename)
    saltmodel, *version = basename.split("_")[1].split("params")[0].split("-")
    saltmodel = saltmodel.strip()
    version = None if len(version) ==0 else version[0].strip()
    if version is None:
        if saltmodel == "salt2": # so name is salt2 only:
            return f"{saltmodel} v=T21" # default version T21
        return saltmodel
    
    return f"{saltmodel} v={version}"

# ================== #
#                    #
#   HOST             #
#                    #
# ================== #
def get_globalhost_data(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_globalhost_prop.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, index_col=0)

def get_localhost_data(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_local2kpc_prop.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, index_col=0)


def get_globalhost_mag():
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/host_prop/host_photometry/ztfdr2_valid_hostphot_mag.csv")
    return pandas.read_csv(filepath, index_col=0)
    

def get_localhost_mag():
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/host_prop/host_photometry/ztfdr2_local2_hostphot_mag.csv")
    return pandas.read_csv(filepath, index_col=0)
    
# ================== #
#                    #
#   LIGHTCURVES      #
#                    #
# ================== #
def get_target_lightcurve(target, test_exist=True, load=True):
    """ """
    fullpath = os.path.join(IDR_PATH, "lightcurves", f"{target}_LC.csv")
    if test_exist:
        if not os.path.isfile(fullpath):
            warnings.warn(f"No lc file for {target} ; {fullpath}")
            return None
        
    if not load:
        return fullpath
    
    return pandas.read_csv(fullpath,  delim_whitespace=True, comment='#')


def get_target_lightcurve_baseline(target):
    """ """
    filepath = get_target_lightcurve(target, load=False)
    lcinfo = open(filepath, "r").read().splitlines()
    i_struct = [i for i, line in enumerate(lcinfo) if line.startswith("# ------")]
    cols, *data = [l.replace("#","").split() for l in lcinfo[i_struct[1]+2: i_struct[2]]]
    return pandas.DataFrame(data=data, columns=cols)


def get_phase_coverage(load=True, warn=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables", "phase_coverage.parquet")
    if not load:
        return filepath

    if not os.path.isfile(filepath):
        if warn:
            warnings.warn(
                "No phase_coverage file. build one using ztfidr.Sample().build_phase_coverage(store=True)")
        return None

    return pandas.read_parquet(filepath, **kwargs)#.set_index(["ztfname", "filter"])


# ================== #
#                    #
#   Spectra          #
#                    #
# ================== #
def get_autotyping(load=True, index_col=0, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/sample_def/spec_class/autotyping.csv")
    if not load:
        return filepath
    return pandas.read_csv(filepath, index_col=index_col, **kwargs)


def get_lightcurve_datafile(contains=None):
    """ """
    from glob import glob
    glob_format = "*" if contains is None else f"*{contains}*"
    files = glob(os.path.join(IDR_PATH, "lightcurves", glob_format))
    datafile = pandas.DataFrame(files, columns=["fullpath"])
    datafile["basename"] = datafile["fullpath"].str.split(pat="/", expand=True).iloc[:, -1]
    datafile["ztfname"] = datafile["basename"].str.split(pat="_", expand=True).iloc[:, 0]
    return datafile
    
def get_spectra_datafile(contains=None, startswith=None,
                         snidres=False, extension=None, use_dask=False,
                         add_phase=True, data=None):
    """ """
    from glob import glob
    glob_format = "*" if not startswith else f"{startswith}*"
    if snidres and extension is None:
        extension = "_snid.h5"
    elif extension is None:
        extension = ".ascii"

    if contains is not None:
        glob_format += f"{contains}*"
    if extension is not None:
        glob_format += f"{extension}"

    specfiles = glob(os.path.join(IDR_PATH, "spectra", glob_format))
    datafile = pandas.DataFrame(specfiles, columns=["fullpath"])
    datafile["basename"] = datafile["fullpath"].str.split(pat="/", expand=True).iloc[:, -1]
    
    specfile = pandas.concat([datafile, parse_filename(datafile["basename"], snidres=snidres)], axis=1)
    
    if add_phase:
        from astropy.time import Time
        if data is None:
            from .sample import get_sample
            data = get_sample().data
            
        specfile["dateiso"] = Time(np.asarray(specfile["date"].apply(lambda x: f"{x[:4]}-{x[4:6]}-{x[6:]}"), dtype=str), format="iso").mjd
        specfile = specfile.join(data[["t0", "redshift"]], on="ztfname")
        specfile["phase_obs"] = (specfile.pop("dateiso")-specfile.pop("t0"))
        specfile["phase"] = specfile["phase_obs"]/(1+specfile.pop("redshift"))
        
    return specfile


def parse_filename(file_s, snidres=False):
    """ file or list of files.
    Returns
    -------
    Serie if single file, DataFrame otherwise
    """

    index = ["ztfname", "date", "telescope", "version"]
    fdata = []
    for file_ in np.atleast_1d(file_s):
        file_ = os.path.basename(file_).split(".ascii")[0]
        if snidres:
            #print(file_)
            name, date, *telescope, origin, snid_ = file_.split("_")
        else:
            try:
                name, date, *telescope, origin = file_.split("_")
            except:
                print(f"failed parsing filename for {file_}")
                continue

        telescope = "_".join(telescope)
        fdata.append([name, date, telescope, origin])

    if len(fdata) == 1:
        return pandas.Series(fdata[0], index=index)
    return pandas.DataFrame(fdata, columns=index)



import numpy as np
from scipy import stats


from astropy.cosmology import Planck18 as cosmo

def get_taylor_hostmass(magi, magg, magi_err, magg_err, redshift, 
                       sigma_int=0.1, cosmo=cosmo,
                       use_prior=False):
    """ computes the Taylor mass from given host magnitudes and redshifts.
    
    If use_prior=True, the g-i color gets a prior to avoid issues with low SNR 
    colors as in Rigault+2020.
    
    sigma_int is is intrinsic scatter of 0.1 dex from Taylor et al
    
    See eq. 8 of Taylor et al 2011 (2011MNRAS.418.1587T)
    ```
      log M∗/[M⊙] = 1.15 + 0.70(g − i) − 0.4Mi
    ```

    Returns
    -------
    mass, mass_err

    """
    color = magg-magi
    color_err = np.sqrt(magg_err**2 + magi_err**2)
    
    if use_prior:
        color, color_err = get_priored_color(color, color_err)

    magabs_i = magi - cosmo.distmod(np.asarray(redshift)).value
    
    mass = 1.15 + 0.70*color - 0.4*(magabs_i)
    mass_err = np.sqrt( (color_err*0.7)**2 + (0.4*magi_err**2) + sigma_int**2)
    
    return mass, mass_err
    
def get_priored_color(c, cerr, ndraw=2000):
    """ """
    c = np.atleast_1d(c)
    cerr = np.atleast_1d(cerr)
    
    xx, priorpdf = gi_prior_pdf(xx=f'-0.5:2.5:{int(ndraw*5)}j')
    likelihood = stats.norm.pdf(xx[:,None], loc=c, scale=cerr)
    posterior = likelihood*priorpdf[:,None] # at a constant
    
    draws = [np.random.choice(xx, size=ndraw, replace=True, p=p_/np.nansum(p_))
         for p_ in posterior.T]

    # assuming gaussian err
    new_c = np.mean(draws, axis=1)
    new_c_err = np.std(draws, axis=1) 
    
    return new_c, new_c_err

from scipy import stats
def gi_prior_pdf(mur=1.25, mub=0.85, sigmar=0.1, sigmab=0.3, b_coef=0.9,
                  xx="-0.5:2.5:1000j"):
    """ """
    if type(xx) == str:
        xx = eval(f"np.r_[{xx}]")
    pdf_b = stats.norm.pdf(xx, loc=mub, scale=sigmab)
    pdf_r = stats.norm.pdf(xx, loc=mur, scale=sigmar)
    return xx, pdf_b * b_coef + pdf_r * (1-b_coef)




import sncosmo
from extinction import ccm89
import numpy as np


def correct_mwextinction_on_hostmag(hostmag, rv=3.1):
    """ """
    hostmag = hostmag.copy() # do not affect the input original
    av = hostmag["mwebv"] * rv
    bands = ["g","r","i","z","y"]
    wave = np.asarray([sncosmo.get_bandpass(f"ps1::{band}").wave_eff for band in bands])
    offsets = np.asarray([ccm89(wave, av_, 3.1) for av_ in av])
    for i, b_ in enumerate(bands):
        hostmag[f"PS1{b_}"] -= offsets[:,i] # reduces the magnitude as it gets brighter

    return hostmag



import pandas
import numpy as np
import warnings


from . import io

TARGET_DATA = io.get_targets_data()

COLORS = ["0.6"]*10 # , "tan", "lightsteelblue", "thistle", "darkseagreen"]

class Target():
    """ """
    def __init__(self, lightcurve, spectra, meta=None):
        """ """
        self.set_lightcurve(lightcurve)
        self.set_spectra(spectra)
        self.set_meta(meta)
        
    @classmethod
    def from_name(cls, targetname, load_snidres=True):
        """ """
        from . import lightcurve, spectroscopy
        lc = lightcurve.LightCurve.from_name(targetname)
        spec = spectroscopy.Spectrum.from_name(targetname, as_spectra=False, load_snidres=load_snidres)
        meta = TARGET_DATA.loc[targetname]
        this = cls(lc, spec, meta=meta)
        this._name = targetname
        return this
        
    
    # -------- #
    #  SETTER  #
    # -------- #    
    def set_lightcurve(self, lightcurve):
        """ """
        self._lightcurve = lightcurve
        
    def set_spectra(self, spectra):
        """ """
        self._spectra = spectra   
    
    def set_meta(self, meta):
        """ """
        self._meta = meta
        
    # -------- #
    #  GETTER  #
    # -------- #
    def get_autotyping(self, full_output=True, remove_fullna=True, allow_snidrun=True):
        """ """
        notyping = pandas.Series([np.NaN]*4, index=['type', 'subtype', 'p(type)', 'p(subtype|type)'])
        if self.nspectra == 0:
            return notyping
                
        types = np.asarray([sn_.get_type() if sn_ is not None else ((np.NaN,np.NaN),(np.NaN, np.NaN))
                            for sn_ in np.atleast_1d(self.get_snidresult(allow_run=allow_snidrun))
                            ]).T
        
        phases = [np.round(spec_.get_phase(self.salt2param['t0']), 1) for spec_ in np.atleast_1d(self.spectra)]
        df_types = pandas.DataFrame(np.concatenate(types,axis=0).T,
                    columns=["type","subtype","p(type)","p(subtype|type)"], 
                    index=pandas.Series(phases, name="phase")).replace({"unclear":np.NaN,"nan":np.NaN})

        
        if remove_fullna:
            df_types = df_types[~df_types.isna().all(axis=1)]

        if full_output:
            return df_types

        df_types = df_types.drop_duplicates()


        if len(df_types)==0:
            return notyping

        if len(df_types)==1:
            return df_types.iloc[0]

        types = df_types["type"].unique()
        if len(types)==1:
            typing = {"type":types[0], "p(type)":df_types["p(type)"].astype(float).max()}
            subtypes = df_types["subtype"].unique()
            if len(subtypes) == 1:
                typing["subtype"] = subtypes[0]
                typing["p(subtype|type)"] = df_types["p(subtype|type)"].astype(float).max()
            else:
                typing["subtype"] = "unclear"
                typing["p(subtype|type)"] = np.NaN

        else:
            typing = {"type":"unclear",
                      "p(type)":np.NaN,
                      "subtype": "unclear",
                      "p(subtype|type)":np.NaN
                     }
        return pandas.Series(typing)[['type', 'subtype', 'p(type)', 'p(subtype|type)']]
        return df_types

    def fit_snid(self, delta_phase=5, lbda_range=[4000, 8000],
                     redshift=None, set_it=True, use_phase=True,
                     **kwargs):
        """ fit SNID constraining redshift and phase """

        # Redshift input
        # Phase
        snidres = []
        for spec_ in np.atleast_1d(self.spectra):
            if use_phase:
                phase = spec_.get_phase( self.meta["t0"] )
            else:
                phase = None
                
            snidres_ = spec_.fit_snid(phase=phase, redshift=redshift,
                                      delta_phase=delta_phase, lbda_range=lbda_range)
            if set_it:
                spec_.set_snidresult(snidres_)
            
            snidres.append(snidres_)
            
        return snidres[0] if not self.has_multiple_spectra() else snidres
        
    
    def get_snidresult(self, redshift=None, zquality=2, set_it=True, allow_run=True, **kwargs):
        """ """
        if not self.has_spectra():
            warnings.warn("No spectra loaded.")
            return None
        
        if not allow_run:
            snidres = [spec_.snidresult for spec_ in np.atleast_1d(self.spectra)]
            return snidres[0] if not self.has_multiple_spectra() else snidres
        
        if redshift is None:
            z_, z_quality_ = self.get_redshift()
            if zquality is not None and \
               zquality not in ["*","all"] and \
               z_quality_ in np.atleast_1d(zquality):
                redshift = z_
            
        snidres = []
        for spec_ in np.atleast_1d(self.spectra):
            phase = spec_.get_phase( self.salt2param["t0"] )
            if spec_.snidresult is None:
                snidres_ = spec_.fit_snid(phase=phase, redshift=redshift, **kwargs)
                if set_it:
                    spec_.set_snidresult(snidres_)
            else:
                snidres_ = spec_.snidresult
                
            snidres.append(snidres_)
                
        return snidres[0] if not self.has_multiple_spectra() else snidres

    def get_redshift(self ):
        """ """
        return self.meta["redshift"], self.meta["source"]

    def get_obsphase(self, **kwargs):
        """ """
        phase_ = self.get_lc_obsphase(**kwargs)
        phase_["spectra"] = self.get_spec_obsphase()
        return phase_
    
    def get_lc_obsphase(self, groupby="filter", min_detection=5, **kwargs):
        """ """
        return self.lightcurve.get_obsphase(groupby=groupby, min_detection=min_detection, **kwargs)

    def get_spec_obsphase(self):
        """ """
        return [spec_.get_phase(self.salt2param["t0"], self.get_redshift()[0])
                                    for spec_ in np.atleast_1d(self.spectra)]
    
    # -------- #
    #  LOADER  #
    # -------- #
    
    # -------- #
    # PLOTTER  #
    # -------- #
    def show(self, nbest=3, lines={"Ha":6563},
                 line_color="0.6", allow_snidrun=False):
        """ """
        import matplotlib.pyplot as mpl
        n_speclines = np.max([1,self.nspectra])
        fig = mpl.figure(figsize=[7.2,3+2.5*n_speclines])
        
        _ = self.get_snidresult(allow_run=allow_snidrun)
        # - Axes
        _lc_height = 0.25/np.sqrt(n_speclines)
        _sp_height = 0.01+0.40/n_speclines
        _spany = 0.04+0.08/np.sqrt(n_speclines)
        _lc_spany = 0.03+0.11/np.sqrt(n_speclines)
        _bottom_lc = 0.04+0.05/np.sqrt(n_speclines)
        _top_lc = _bottom_lc+_lc_height

        axlc = fig.add_axes([0.15, _bottom_lc, 0.78, _lc_height])
        axes = []
        for i in range(n_speclines):
            axs = fig.add_axes([0.15, _top_lc+_lc_spany+i*(_spany+_sp_height), 0.55, _sp_height])
            
            axt = fig.add_axes([0.75, _top_lc+_lc_spany+i*(_spany+_sp_height), 0.2, _sp_height*0.8])
            axtt = fig.add_axes([0.75, _top_lc+_lc_spany+i*(_spany+_sp_height) + _sp_height*0.9, 0.2, _sp_height*0.1])
            axes.append([axs, [axtt,axt]])

        #
        # - Plotter
        #
        # LightCurves
        lc = self.lightcurve.show(ax=axlc, 
                                  zprop=dict(ls="-", color="0.6",lw=0.5))
        redshift, redshift_source = self.get_redshift()
        redshift_label = redshift_source
        
        # - No Spectra
        if not self.has_spectra():
            axs.text(0.5,0.5, f"no Spectra for {self.name}", 
                         transform=axs.transAxes,
                         va="center", ha="center")
            axs.set_yticks([]) ;axs.set_xticks([])
            clearwhich = ["left","right","top"] # "bottom"
            [axs.spines[which].set_visible(False) for which in clearwhich]
            
        # - spectra            
        else:
            for i, spec_ in enumerate(np.atleast_1d(self.spectra)[::-1]):
                phase = spec_.get_phase( self.salt2param["t0"], z=redshift)

                label=rf"{self.name} z={redshift:.3f} ({redshift_label}) | $\Delta$t: {phase:+.1f}"
                sp = spec_.show_snidresult(axes=axes[i], nbest=nbest,
                                          label=label, color_data=COLORS[i])
                # - ObsLine
                axlc.axvline(spec_.get_obsdate().datetime, 
                             ls="--", color=COLORS[i])
                if i>0:
                    axes[i][0].set_xlabel("")
                else:
                    axes[i][0].set_xlabel(axes[i][0].get_xlabel(), fontsize="medium")

                if lines is not None:
                    for line, lbda in lines.items():
                        for lbda_ in np.atleast_1d(lbda):
                            axes[i][0].axvline(lbda_*(1+redshift), ls="--",
                                           color=line_color, zorder=1, lw=0.5, alpha=0.8)

        return fig
    # ================ #
    #    Properties    #
    # ================ #
    @property
    def lightcurve(self):
        """ """
        return self._lightcurve
    
    @property
    def spectra(self):
        """ """
        return self._spectra
    
    def has_spectra(self):
        """ """
        return self.spectra is not None
    
    def has_multiple_spectra(self):
        """ """
        return self.nspectra>1
    
    @property
    def nspectra(self):
        """ """
        if not self.has_spectra():
            return 0
        return len(np.atleast_1d(self.spectra))
    
    @property
    def meta(self):
        """ """
        return self._meta
    
    @property
    def salt2param(self):
        """ shortcut to self.lightcurve.salt2param"""
        return self.lightcurve.salt2param
    
    @property
    def name(self):
        """ """
        if not hasattr(self, "_name"):
            return self.meta.name
        return self._name
    





from . import io

targets = io.get_target_data()


def run_snid(filepath,
             lbda_range=[4000,9000],
             delta_redshift=0.05, verbose=False,
             **kwargs):
    """ """
    from .spectroscopy import Spectrum
    spec = Spectrum.from_filename(filepath)
    t0, z = targets.loc[spec.targetname][["t0","redshift"]].values
    phase = spec.get_phase(t0, z=z)
    #if redshift quality is 2, delta_Redshift could be reduced
    return spec.fit_snid(phase=phase, lbda_range=lbda_range, 
                         redshift=z, delta_redshift=delta_redshift,
                         verbose=verbose, **kwargs)



import pandas
import numpy as np
import matplotlib.pyplot as plt

from . import utils

# ================== #
#                    #
#   SAMPLE           #
#                    #
# ================== #
def show_hubble_standardisation(sample_or_data, fig=None,
                                param="c", vmin=-0.2, vmax=0.3, cmap="coolwarm", 
                                clear_axes=False, bins="auto",
                                zref = 0.204, lw=2, rotation=9):
    """ """
    from astropy.cosmology import Planck18
    
    if not type(sample_or_data) == pandas.DataFrame:
        data = sample_or_data.get_data(x1_range=[-4,4], c_range=[-0.3,0.8], goodcoverage=True)
        data = data[data["classification"].isin(['snia-norm'])]
    else:
        data = sample_or_data.copy()
        
    if fig is None:
        fig = plt.figure(figsize=[6,4])
    
    ax = fig.add_axes([0.1,0.15,0.8,0.75])
    cax = fig.add_axes([0.4,0.3,0.4,0.25])

    if "mag" not in data:
        data["mag"] = -2.5 * np.log10( data["x0"] ) + 19*1.58
        
    sc = ax.scatter(data["redshift"], 
                    data["mag"], c=data[param], 
                    cmap=cmap,
                    s=10, vmin=vmin, vmax=vmax)
    
    _ = utils.hist_colorbar(data[param], ax=cax, cmap=cmap, fcolorbar=0.1,
                           vmin=vmin, vmax=vmax, bins=bins)

    xx = np.linspace(0.00001,0.4,1000)
    ax.plot(xx, Planck18.distmod(xx), color="k", lw=lw)

    ax.set_ylabel("distance modulus + cst", fontsize="large")
    ax.set_xlabel("redshift", fontsize="large")
    cax.set_xlabel(f"{param}", fontsize="small", color="0.5")
    cax.tick_params(axis="x", labelsize="small", 
                   labelcolor="0.5", color="0.5")

    ax.text(zref, Planck18.distmod(zref).value, "Planck H(z)", va="center", ha="center", 
           fontsize="x-small", rotation=rotation, weight="bold",
            bbox={"facecolor":"w", "edgecolor":"None"})
    
    if clear_axes:
        clearwhich = ["bottom","right","top","left"] # "bottom"    
        [ax.spines[which].set_visible(False) for which in clearwhich]
        ax.tick_params(labelsize="medium")
#    ax.set_xticks([])

    return fig

# ================== #
#                    #
#   Lightcurves      #
#                    #
# ================== #
def show_npoints_distribution_perband(sample_or_phase_coverage, ax=None, phase_coverage=None, 
                                      clearaxes=True, add_pantheon=True):
    """ """
    if type(sample_or_phase_coverage) is pandas.DataFrame:
        phase_coverage = sample_or_phase_coverage
    else:
        phase_coverage = sample_or_phase_coverage.get_phase_coverage()


    from matplotlib.colors import to_rgba
    if ax is None:
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=[6, 3.5])
        axb = fig.add_axes([0.08, 0.15, 0.87, 0.35])
        axt = fig.add_axes([0.08, 0.6, 0.87, 0.35])

    else:
        fig = ax.figure
        axb, axt = fig.axew

    prop = dict(bins=np.logspace(0,3, 15), density=False, log=False)

    # -------------
    weights = np.ones(len(phase_coverage))*1
    axt.hist(phase_coverage["n_early_points_ztfg"]+phase_coverage["n_late_points_ztfg"], histtype="step", 
            fill=True, 
            edgecolor=to_rgba("tab:green", 1), lw=1., ls="--",
            facecolor=to_rgba("tab:green", 0.05), 
            label=f"ztf:g", weights=weights,
            zorder=2, **prop)

    axt.hist(phase_coverage["n_early_points_ztfr"]+phase_coverage["n_late_points_ztfr"], histtype="step", 
            fill=True, 
            edgecolor=to_rgba("tab:red", 1), lw=1., ls="-",
            facecolor=to_rgba("tab:red", 0.05), 
            label=f"ztf:r",weights=weights,
            zorder=3, **prop)

    axt.hist(phase_coverage["n_early_points_ztfi"]+phase_coverage["n_late_points_ztfi"], histtype="step", 
            fill=True, 
            edgecolor=to_rgba("tab:orange", 1), lw=1., ls=":",
            facecolor=to_rgba("tab:orange", 0.05), 
            label=f"ztf:i",weights=weights,
            zorder=5, **prop)


    # --------------
    weights = np.ones(len(phase_coverage))*1

    axb.hist(phase_coverage["n_points"], histtype="step", color="k", lw=1.5,
            label=f"any phase",
            weights=weights,
            zorder=5, **prop)

    axb.hist(phase_coverage["n_early_points"], histtype="step", 
            fill=True, 
            edgecolor=to_rgba("0.5", 1), lw=1., ls="-",
            facecolor=to_rgba("0.5", 0.), 
            label="pre-max",
            weights=weights,            
            zorder=3, **prop)

    axb.hist(phase_coverage["n_late_points"], histtype="step", 
            fill=True, 
            edgecolor=to_rgba("0.5", 1), lw=0., 
            facecolor=to_rgba("0.5", 0.3), 
            label="post-max",
            weights=weights,
            zorder=4, **prop)

    # --------------

    axb.set_xlabel("number of detected point", fontsize="large")
    axb.set_xscale("log")
    axt.set_xscale("log")

    if clearaxes:
        clearwhich = ["left","right","top"] # "bottom"
        for ax in [axb, axt]:
            [ax.spines[which].set_visible(False) for which in clearwhich]
            ax.tick_params(axis="y", labelsize="small", 
                       labelcolor="0.5", color="0.5")

    axt.legend(fontsize="medium", frameon=False)#, loc="upper left")
    axb.legend(fontsize="medium", frameon=False)

    axt.set_xlim(*axb.get_xlim())
    return fig


def show_firstdet_distributions(sample, ax=None, restrict_to=[-25,60]):
    """ """
    phases = sample.get_phases(phase_range=restrict_to)
    goodlc = sample.get_data(goodcoverage=True, 
                             x1_range=[-4,4], 
                             c_range=[-0.3,0.8],
                             t0_err_range=[0,1],
                             x1_err_range=[0,1], 
                             c_err_range=[0,0.1],
                            ).index.astype(str)

    from matplotlib.colors import to_rgba
    if ax is None:
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=[6, 2.5])
        ax = fig.add_axes([0.08, 0.1, 0.87, 0.8])
    else:
        fig = ax.figure

    first_det_per_band = phases.groupby(level=[0,1]).min()
    first_det = first_det_per_band.unstack().min(axis=1)

    prop = dict(range=[-21,30], bins=50, histtype="step", fill=True)

    # ---- Any band

    _ = ax.hist(first_det, facecolor="None", zorder=5,
                color="0.5", label="all targets", **prop)

    _ = ax.hist(first_det.loc[goodlc], zorder=4,
                facecolor="darkkhaki", lw=0,
                 label="‘Basic cuts’", **prop)
    # ---- Per filter

    #ax.legend(frameon=False)

    for band, color, label in zip(["ztfg","ztfr","ztfi"],
                          ["tab:green","tab:red","tab:orange"],
                          ["ztf:g","ztf:r","ztf:i"]):
        band_fdet = first_det_per_band.xs(band, level=1)
        band_fdet_good = band_fdet.loc[band_fdet.index.isin(goodlc)]


        _ = ax.hist(band_fdet_good, 
                    facecolor=to_rgba(color, 0.0), #zorder=2,
                    edgecolor=color,
                    weights = np.ones(len(band_fdet_good))*-1, 
                    label=label, **prop)

    ax.spines["bottom"].set_position(('data',0))
    ax.spines["bottom"].set_zorder(10)
    yticks = ax.get_yticks()
    ax.set_yticks(yticks, labels=[f"{int(np.abs(y_))}" for y_ in yticks])
    
    
    clearwhich = ["left","right","top"] # "bottom"
    [ax.spines[which].set_visible(False) for which in clearwhich]
    ax.tick_params(axis="y", labelsize="small", 
               labelcolor="0.5", color="0.5")
    ax.tick_params(axis="x", pad=2, zorder=10)
    ax.set_xlabel("First detection phase [days]", loc="right")


    #ax.legend(["all", "good"], frameon=False)
    ax.legend(frameon=False, ncol=3)
    return fig


from pprint import pprint
import numpy as np


def show_residual(data, key="pull", ax=None, axh=None, axc=None,
                  binstat="mean", bins=np.arange(-50,50, 0.25),
                  main_color="C0", cmap_significance="GnBu_r",
                  incl_lef=True, incl_runningpull=False,
                  pull_color=None,
                  blw=0.1, bs=8, bec="k",
                  **kwargs):
    """ """
    from matplotlib.colors import to_rgba
    from scipy import stats
    
    if ax is None:
        fig = plt.figure(figsize=[7,3])
        ax = fig.add_axes([0.1,0.2,0.7,0.7])
        axh = fig.add_axes([0.801,0.2,0.1,0.7])
        if cmap_significance is not None:
            axc =  fig.add_axes([0.801,0.9,0.1,0.02])
    else:
        fig = ax.figure

    prop = dict(s=0.2, lw=0, facecolors=main_color, edgecolor="None", zorder=2)
        
    data = data.copy() # not changing the input dataframe
    ax.scatter(data["phase"], data[key],**{**prop,**kwargs})
    if axh is not None:
        axh.hist(data[key], orientation="horizontal", range=[-6,6], bins="auto",
                 density=True, 
                 facecolor=to_rgba(main_color, 0.5),
                 edgecolor=to_rgba(main_color, 0.8),
                 histtype="step", fill=True, lw=1, zorder=1)
        axh.set_xticks([])
        axh.set_yticks([])
    
    lbins = 0.5*(bins[1:] + bins[:-1])
    data["phasebins"] = pandas.cut(data["phase"], bins, labels=lbins)
    bin_stats = data.groupby("phasebins")[key].agg([binstat, "std", "size", utils.nmad])
    
    
    prop_binstat = dict(s=bs, lw=blw, zorder=8, edgecolor=bec)
    if cmap_significance is not None:
        cmap = plt.get_cmap(cmap_significance, 5)
        significance = bin_stats[binstat] / (bin_stats["std"]/np.sqrt(bin_stats["size"]-1))
        if incl_lef:
            print("include look elsewhere effect")
            ntrials = len(lbins)
            proba = stats.norm.sf(np.abs(significance.values), loc=0, scale=1)*2 # x2 because two sides the same
            proba[proba<1e-15] = 1e-15
            proba_lef = np.asarray( 1- (1-proba)**ntrials, dtype="float64")
            significance = np.abs(-stats.norm.ppf(proba_lef/2)) # abs to avoid -0
            bin_stats["significance"] = significance
        else:
            bin_stats["significance"] = significance.abs()
            
        prop_binstat = {**prop_binstat, 
                        **dict(c=bin_stats["significance"], vmax=5, vmin=0, cmap=cmap)
                       }
        
    else:
        cmap = None
        prop_binstat = {**prop_binstat, 
                        **dict(facecolors="w")
                       }
        
    sc = ax.scatter(bin_stats.index, bin_stats[binstat], 
               **prop_binstat)

    if incl_runningpull:
        if pull_color is None:
            pull_color = to_rgba("w",0.3)
        ax.fill_between(bin_stats.index, bin_stats["nmad"], -bin_stats["nmad"],
                            color=pull_color, lw=1, zorder=5)
        
    if cmap is not None and axc is not None:
        fig.colorbar(sc, cax=axc, orientation="horizontal")

    if key in ["pull","pull_corr"]:
        ax.axhspan(-1,1, color=main_color, alpha=0.05, zorder=1)
        ax.axhspan(-2,2, color=main_color, alpha=0.05, zorder=1)
        ax.axhspan(-3,3, color=main_color, alpha=0.05, zorder=1)
        ax.set_ylim(-6,6)
        if axh is not None:
            from scipy import stats 
            # pull hist
            xx = np.linspace(-6,6, 100)
            axh.plot(stats.norm.pdf(xx, loc=0, scale=1), 
                     xx, lw=0.5, color="k", ls="--")
#            axh.plot(stats.norm.pdf(xx, loc=0, scale=1.2), 
#                     xx, lw=1, color="k", ls="-")
            
    ax.set_xlim(-50,50)
    ax.axhline(0, color="k", lw=1)
    ax.set_xlabel("phase [in days]", fontsize="large")
    ax.set_ylabel(f"pull [$\sigma$]", fontsize="large")
    if axh is not None:
        axh.set_ylim(*ax.get_ylim())
        clearwhich = ["bottom","right","top","left"] # "bottom"    
        [axh.spines[which].set_visible(False) for which in clearwhich]

    return fig

def show_lc_residuals(data, fig=None, axes=None,
                      incl_lef=True, cmap_significance="GnBu_r", 
                      binstat="mean", key="pull",
                      bincadence=[1,1,2], # g, r, i
                      incl_runningpull=True, incl_filtername=True,
                      incl_linelabel=True, 
                      scoef_i=3, incl_hist=True,
                      phase_lines = [-20, -15, -10, -5, 0, 10, 20, 30, 40, 50],
                      propaxes={},
                        **kwargs):
    """ """
    from matplotlib.colors import to_rgba
    # data
    data_g = data[data["filter"]=="ztfg"]
    data_r = data[data["filter"]=="ztfr"]
    data_i = data[data["filter"]=="ztfi"]

    
    left = propaxes.get("left", 0.1)
    bottom, height = propaxes.get("bottom", 0.1), propaxes.get("height", 0.27)
    vspan, hspan = propaxes.get("vspan", 0.015),  propaxes.get("hspan", 0.001)
    width, hwidth = propaxes.get("width", 0.78),  propaxes.get("hwidth", 0.1)
    cvspan = propaxes.get("cvspan", 0.)
    print(cvspan)
    # -- Figure layout -- #
    if axes is None:
        if fig is None:
            fig = plt.figure(figsize=[7,5])
            
        # axes
        axg = fig.add_axes([left, bottom+2*(height+vspan), width, height], zorder=3)
        #
        axr = fig.add_axes([left, bottom+1*(height+vspan), width, height], zorder=3)
        #
        axi = fig.add_axes([left, bottom+0*(height+vspan), width, height], zorder=3)

        if incl_hist:
            ahg = fig.add_axes([left+width+hspan, bottom+2*(height+vspan), hwidth, height], zorder=1)    
            ahr = fig.add_axes([left+width+hspan, bottom+1*(height+vspan), hwidth, height], zorder=1)
            ahi = fig.add_axes([left+width+hspan, bottom+0*(height+vspan), hwidth, height], zorder=1)
        else:
            ahg, ahr, ahi = None, None, None

        axc = fig.add_axes([left, bottom+3*(height+vspan)+cvspan, 0.1, 0.01], zorder=3)
        
    else:
        if len(axes) == 3:
            axg, axr, axi = axes
            ahg, ahr, ahi = None, None, None
            axc = None
        elif len(axes) == 4:
            axg, axr, axi, axc = axes
            ahg, ahr, ahi = None, None, None
        elif len(axes) == 6:
            axg, axr, axi, ahg, ahr, ahi = axes
            axc = None
        elif len(axes) == 7:
            axg, axr, axi, ahg, ahr, ahi, axc = axes
        else:
            raise ValueError(f"cannot parse the input axes size {len(axes)}, 3, 4, 6 or 7 expected.")
        
        fig = axg.figure
        
    # 


    # -- plotting -- #

    prop = {**dict(binstat=binstat, bs=20, blw=0.2, incl_lef=incl_lef,
               cmap_significance=cmap_significance, key=key,
                       incl_runningpull=incl_runningpull), **kwargs}

    _ = show_residual(data_g, ax=axg, axh=ahg, main_color="tab:green", 
                      bins=np.arange(-50,50, bincadence[0]), 
                      axc = axc, pull_color=to_rgba("w",0.4),
                      **prop
                     )
    _ = show_residual(data_r, ax=axr, axh=ahr,main_color="tab:red", 
                      bins=np.arange(-50,50, bincadence[1]), pull_color=to_rgba("w",0.4),
                      **prop
                     )

    _ = show_residual(data_i, ax=axi, axh=ahi, main_color="tab:orange",  
                      bins=np.arange(-50,50, bincadence[2]),
                      s=prop.pop("s",1)*scoef_i, pull_color=to_rgba("w",0.55),
                      bs=prop.pop("bs",8)*1.2,
                      **prop)

    # -- texts -- #
    if incl_filtername:
        proptext = dict(rotation=-90, va="center", ha="left", zorder=1)
        axg.text(1.005,.5, "ztf:g", color="tab:green", transform=axg.transAxes, **proptext) 
        axr.text(1.005,.5, "ztf:r", color="tab:red", transform=axr.transAxes, **proptext)
        axi.text(1.005,.5, "ztf:i", color="tab:orange", transform=axi.transAxes, **proptext)
    
    
    [ax_.set_yticks([-4,-2,0,2,4]) for ax_ in [axg, axr, axi]]
    # -- phase lines -- #
    if phase_lines is not None:
        [ax_.axvline(l, lw=0.5, color="0.8", zorder=1) for ax_ in [axg, axr, axi] for l in phase_lines]

        if incl_linelabel:
            _ = [axg.text(l, 5., l, va="bottom", ha="center", color="0.7", fontsize="small") for l in phase_lines]
    
    # -- layout -- #
    _ = [ax_.set_ylim(-5,5) for ax_ in fig.axes if ax_ is not axc]
    _ = [ax.tick_params(labelsize="medium") for ax in fig.axes]
    _ = [ax_.set_xticklabels([]) for ax_ in [axr, axg]]
    
    if axc is not None:
        axc.set_xticks([])
        prop_clabel = dict(fontsize="small", transform=axc.transAxes,  color="0.5")
        axc.text(-0.01, 0.5, "0 ", va="center", ha="right", **prop_clabel)
        axc.text(1.01,0.5, " 5", va="center", ha="left", **prop_clabel)
        axc.text(0.5,1.01, "significance", va="bottom", ha="center", **prop_clabel)
        
    return fig



# ================== #
#                    #
#   Spectra          #
#                    #
# ================== #
def show_peakmag_barhist(sample_or_data, bts_data=None,
                            fig=None,
                            bins=np.linspace(12,22, 41),
                            facecolor_color="w", main_color="0.7",
                            filtername="g", xlim=[14,21]
                        ):
    """ 
    """
    if bts_data is None:
        bts_data = pandas.read_csv("https://sites.astro.caltech.edu/ztf/bts/explorer.php?f=s&subsample=all&classstring=&classexclude=&ztflink=lasair&lastdet=&startsavedate=&startpeakdate=&startra=&startdec=&startz=&startdur=&startrise=&startfade=&startpeakmag=&startabsmag=&starthostabs=&starthostcol=&startb=&startav=&endsavedate=&endpeakdate=&endra=&enddec=&endz=&enddur=&endrise=&endfade=&endpeakmag=19.0&endabsmag=&endhostabs=&endhostcol=&endb=&endav=&format=csv")
        
    bts_data = bts_data[bts_data["type"] != "-"] # secured typing in BTS
    
    if not type(sample_or_data) == pandas.DataFrame:
        data = sample_or_data.get_data()
    else:
        data = sample_or_data
    
    # Figure
    if fig is None:
        fig = plt.figure(figsize=[7,3])
    
    ax = fig.add_axes([0.10,0.15,0.8,0.75])
    
    data_inbts = data.loc[data.index.isin(bts_data["ZTFID"])]
    ampl, _ = np.histogram(data[f"peak_mag_ztf{filtername}"], bins)
    ampl_inbts, _ = np.histogram(data_inbts[f"peak_mag_ztf{filtername}"], bins)

    # DR2 hist
    ax.bar(bins[:-1], ampl, width=0.42/2, align="edge", 
           facecolor=facecolor_color, 
           edgecolor=main_color, 
           label=f"ZTF Cosmo DR2")

    # DR2 in BTS hist
    ax.bar(bins[:-1], ampl_inbts, width=0.42/2, align="edge", 
           facecolor=main_color, 
           edgecolor="None", 
           label=f"in BTS")

    ax.set_xticks(np.arange(bins[0], bins[-1], 1))

    ax.set_xlim(*xlim)
    ax.set_xlabel(f"SALT2 peak magnitude [ztf:{filtername}]", fontsize="medium")

    clearwhich = ["right","top", "left"] # "bottom"
    [ax.spines[which].set_visible(False) for which in clearwhich]
    ax.tick_params("y", labelsize="small")
    ax.legend(loc="upper left", frameon=False, ncol=1, fontsize="medium")

    return fig


def show_spectra_sources(specdata=None, bts_data=None, min_nspec=50, 
                            main_color = "0.7", facecolor_color = "w"):
    """ """
    # spectrum files    
    if specdata is None:
        from . import io
        specdata = io.get_spectra_datafile()
        
    # BTS data    
    if bts_data is None:
        bts_data = pandas.read_csv("https://sites.astro.caltech.edu/ztf/bts/explorer.php?f=s&subsample=all&classstring=&classexclude=&ztflink=lasair&lastdet=&startsavedate=&startpeakdate=&startra=&startdec=&startz=&startdur=&startrise=&startfade=&startpeakmag=&startabsmag=&starthostabs=&starthostcol=&startb=&startav=&endsavedate=&endpeakdate=&endra=&enddec=&endz=&enddur=&endrise=&endfade=&endpeakmag=19.0&endabsmag=&endhostabs=&endhostcol=&endb=&endav=&format=csv")
        
    bts_data = bts_data[bts_data["type"] != "-"] # secured typing in BTS
    # spectra in BTS
    specdata_inbts = specdata[ specdata["ztfname"].isin(bts_data["ZTFID"]) ]
    
    # local function
    def get_specsource_distribution(data_, min_number=min_nspec):
        df = data_.copy()
        ndata = df.groupby("telescope").size().sort_values(ascending=True)#.reset_index()
        as_others = ndata[ndata<=min_number].index
        df.loc[df["telescope"].isin(as_others),"telescope"] = "Other"
        df[df["telescope"].isin(as_others)]
        return df.groupby("telescope").size().sort_values(ascending=True).reset_index()
        #host_redshifts[host_redshifts["telescope_source"].isin(as_others)]["telescope_source"]
    
    
    
    ndatas = get_specsource_distribution(specdata).rename({0:"dr2"}, axis=1)
    ndatas_inbts = get_specsource_distribution(specdata_inbts).rename({0:"bts"}, axis=1)
    ndatas = ndatas.merge(ndatas_inbts, on="telescope", how="left").fillna(0)


    fig = plt.figure(figsize=[7,4])
    ax = fig.add_axes([0.12,0.15,0.8,0.75])

    hbars = ax.barh(ndatas.index, ndatas["dr2"], tick_label=ndatas["telescope"],
                   facecolor=facecolor_color, 
                    edgecolor=main_color, label="ZTF Cosmo DR2")

    hbars_bts = ax.barh(ndatas.index, ndatas["bts"], tick_label=ndatas["telescope"],
                   facecolor=main_color, edgecolor=main_color, label="in BTS")


    ax.tick_params("y", labelsize="small", color=main_color, length=0)
    ax.bar_label(hbars, fmt='%d', color=main_color, padding=2, fontsize="small")

    ax.legend(loc="lower right", frameon = False, ncol=1, fontsize="medium")
    clearwhich = ["bottom","right","top", "left"] # "bottom"
    [ax.spines[which].set_visible(False) for which in clearwhich]
    ax.set_xticks([])
    
    return fig


def show_phase_hist(specdata, sample_or_data=None, 
                    main_color = "darkkhaki",
                    second_color = None, xrange = [-22, 60], **kwargs):
    """ 
    """
    if second_color is None:
        second_color = main_color
        
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=[7,2])
    ax = fig.add_axes([0.1,0.25,0.8,0.7])

    if sample_or_data is None:
        from . import get_sample
        sample_or_data = get_sample()

    
    if type(sample_or_data) == pandas.DataFrame:
        good_targets = sample_or_data.index
    else:
        good_targets = sample_or_data.get_data(goodcoverage=True, x1_range=[-4,4], c_range=[-0.3,0.8]).index


    prop = {**dict(bins=41, range=xrange, histtype="step", fill=True),
            **kwargs}
    
    _ = ax.hist(specdata["phase"], facecolor="None", edgecolor=main_color, 
                label="all targets", zorder=3,
                **prop)
    _ = ax.hist(specdata.query("ztfname in @good_targets")["phase"], 
                facecolor=second_color, edgecolor="None",
                label="after 'Basic cuts'",zorder=2,
                **prop)

    ax.set_xlim(*xrange)
    ax.set_xlabel("Phase [days]")
    ax.tick_params("y", labelsize="small", color="0.7", labelcolor="0.7")

    ax.legend(loc="upper right", frameon = False, ncol=1, fontsize="medium")
    clearwhich = ["right","top", "left"] # "bottom"
    [ax.spines[which].set_visible(False) for which in clearwhich]
    
    return fig


# ================== #
#                    #
#   Typings          #
#                    #
# ================== #

def show_typingdistribution(sample, ax=None, fig=None):
    """ """
    dist_typing = sample.data.groupby("classification").size()
    ia_norm = dist_typing[["snia-norm"]].sum()
    ia_pec = dist_typing[["snia-pec-91t","snia-pec-91bg","snia-pec"]].sum()
    ia = dist_typing[["snia"]].sum()
    rest = dist_typing.sum() - (ia_norm+ia_pec+ia)
    if ax is None:
        if fig is None:
            fig = plt.figure(figsize=[7,3])
        ax = fig.add_subplot(111)
    else:
        fig = ax.figure

    ax.barh(1, ia_norm, facecolor="C0", edgecolor="None")
    ax.barh(0, ia, facecolor="0.7", edgecolor="None")
    ax.barh(-1, ia_pec, facecolor="C1", edgecolor="None")
    ax.barh(-2, rest, facecolor="None", edgecolor="k")
        
    ax.text(ia_norm+30,1, f'{ia_norm}', fontsize="large",
            color="C0", va="center", ha="left", weight="bold")
    ax.text(ia+30,0, f'{ia}', fontsize="large",
            color="0.7", va="center", ha="left", weight="bold")
    ax.text(ia_pec+30,-1, f'{ia_pec}', fontsize="large",
            color="C1", va="center", ha="left", weight="bold")

    ax.text(rest+30,-2, f'{rest}', fontsize="large",
            color="k", va="center", ha="left")

    clearwhich = ["bottom","right","top",] # "bottom"
    [ax.spines[which].set_visible(False) for which in clearwhich]
    ax.set_xticks([])
    ax.set_yticks([1,0,-1,-2], ["SN Ia\nnorm","SN Ia","SN Ia\npeculiar", "Unclear\nnon-ia"])

    return fig

def show_typingorigin(typings):
    """ """    
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=[7,3])
    ax = fig.add_axes([0.15,0.2, 0.7,0.75])


    color=["tab:blue","beige", "0.8"]

    joined = typings.groupby(level=0).sum().sort_values(ascending=False)
    labels = joined.index.str.replace("pec-","pec\n")

    for i, category in enumerate(joined.index):
        d_ = typings.xs(category)[["auto","master","arbiter"]].cumsum()[::-1]
        label = d_.index
        _ = ax.bar(labels[i], d_, color=color,
                  label=None if i>0 else label)

    cbar = ax.bar(labels, joined.values, facecolor="None", edgecolor="k", lw=1)
    ax.bar_label(cbar, fmt='%d', color="k", padding=2, weight="bold", fontsize="small")

    clearwhich = ["right","top","left"] # "bottom"    
    [ax.spines[which].set_visible(False) for which in clearwhich]
    ax.tick_params(labelsize="large")

    _ = ax.set_yticks([])
    _ = ax.set_yscale("log")

    ax.legend(ncol=1, frameon=False, loc="upper right")
    return fig




# basics
import os
import numpy as np
import pandas
from scipy.stats import median_abs_deviation

# special
from pysnid import snid

# in package
from .spectroscopy import Spectrum


SNIDPARSE_TYPING = {"snia": "Ia",
                    "snia-norm": "Ia-norm",
                    "snia-pec-91t": "Ia-91T",
                    "snia-pec-91bg": "Ia-91bg",
                    "snia-pec": "!Ia-norm,Ia-91T,Ia-91bg",
                    }

INSTRUMENT_ZOFFSET = pandas.Series({"SEDm":-0.0010}, name="zoffset")
INSTRUMENT_ZOFFSET.index.name = "instrument"


# Special cases
RM_TYPING_REDSHIFT = ["ZTF19abcttsc", "ZTF19abzlsbl", "ZTF19adajqwl", "ZTF20abgfekk", "ZTF20accmutv", 
                     "ZTF20acyroke", "ZTF18absbspk",
                     "ZTF18abhhxcp","ZTF19aahsclk", "ZTF19abzhvxk" # GAL
                      # rest don't need snidauto
                      "ZTF18abhpgje", "ZTF18acaipdc", "ZTF18acajkff", "ZTF20acvbrbv"
                      # Redshift gain by rm
                      "ZTF18aabxrjp",
                      # redshift
                      "ZTF18aahshhp","ZTF18acefgoc", "ZTF18adatosi", "ZTF19abaukyt", "ZTF19acxngol",
                      "ZTF20aaurhzc", "ZTF20abrjmgi", "ZTF20absitlr", "ZTF20ackitai", "ZTF20acquetr",
                      "ZTF20acxdawc", "ZTF20aciwcuz",
                       #  Gal
                      "ZTF18aaisqmw", "ZTF18abuatfp", "ZTF18accvyao", "ZTF18acpuxuh", "ZTF18aczdapb", "ZTF19adcetym",
                      "ZTF20aanaeev", "ZTF20aankixb", "ZTF20acqikeh", "ZTF20acywbes"
                     ]
    
FORCE_SNIA = ["ZTF18actuhrs"]
#
#
#
#
#

def run_snid(spec_or_filepath, redshift=None, t0=None, t0_err=None,
                delta_redshift=0.05,
                delta_phase=4, # for phase
                lbda_range = [3_500, 10_000], 
                rm_spec_edgepixel = 10, verbose=False,
                **kwargs):
    """ """
    if type(spec_or_filepath) is str:
        spectrum = Spectrum.from_filename(spec_or_filepath)
    else:
        spectrum = spec_or_filepath
    
    # update the lbda_range to be the most conservative between spectrum and input lbda_range
    if lbda_range is not None:
        lbda_range = lbda_range.copy()
        lbda_range[0] = np.max([lbda_range[0], int(spectrum.lbda[rm_spec_edgepixel]+1)])
        lbda_range[1] = np.min([lbda_range[1], int(spectrum.lbda[-rm_spec_edgepixel]-1)])
        print(lbda_range)
    
    if t0 is not None:
        phase = spectrum.get_phase(t0, z=redshift, from_target=False) # take the data we give you
    else:
        phase = None
         
            
    if t0_err is not None:
        delta_phase = np.sqrt(delta_phase**2 + t0_err**2)        
        
    if phase is not None:
        print(f"using phase: {phase} ± {delta_phase:.3f}")
        
    return spectrum.fit_snid(lbda_range=lbda_range, 
                            phase=phase, 
                            redshift=redshift, 
                            delta_redshift=delta_redshift, 
                            delta_phase=delta_phase,
                            verbose=verbose, **kwargs)
#
#
#
#
#
def specfiles_to_snidfiles(filepath):
    """ convert spectrum filename to a snid fit filename
    
    Parameters
    ----------
    filepath: str
        path file name
        
    Returns
    -------
    str
    """
    return filepath.replace(".ascii", "_snid.h5")

def parse_filename(filename):
    """ """
    basename = os.path.basename(filename).split(".")[0]        
    to_parse = basename.split("_")[:4]
    return dict(zip(["targetname", "date", "instrument", "version"],to_parse))

def read_target_snidres(ztfname, instrument="*", concat=True, verbose=False):
    """ reads snid.h5 associated to the target name
    
    Parameters
    ----------
    ztfname: str
        name of the ZTF target
        
    Returns
    -------
    DataFrame
        concatenated snid results dataframe 
    """
    from glob import glob
    from .io import IDR_PATH
    
    if instrument is None:
        instrument = "*"
    elif type(instrument) != str: # assumed list
        instrument = f"[{','.join(instrument)}]*"
    
    snid_target = glob( os.path.join(IDR_PATH, "spectra", f"{ztfname}*_{instrument}_*.h5") )
    if verbose:
        print(snid_target)
        
    sres = [snid.SNIDReader.from_filename(filename).results 
            for filename in snid_target]
    keys = [parse_filename(filename)["instrument"] for filename in snid_target]
    if len(sres) < 1 :
        return None
    
    if not concat:
        return sres
    
    return pandas.concat(sres, keys=keys).reset_index(names=["instrument","no."])

def get_target_snidres(ztfname, 
                        typing=None, rlap_range=[5, None], 
                        n_range=None, grade="good",
                        verbose=False, instrument="*",
                        **kwargs):
    """ get all snid results associated to the target.
    
    Parameters
    ----------
    ztfname: str
        name of the ZTF target

    typing: str
        classification of the target. 
        This is used to select considered snid results entries
        
    rlap_range: None, list
        if not None: [min, max] of the rlap. 
        Use None for no limit, e.g., [5, None] means greater than 5.
        
    n_range: None, list
        if not None: [min, max] of the entry no. 
        Use None for no limit, e.g., [None, 10] means first 10th entries.
        
    grade: None, str, list
        select the grade to considered. If None, no cut.
        
    Returns
    -------
    DataFrame
        concatenated snid results dataframe.
    """
    snidresults = read_target_snidres(ztfname, instrument=instrument)
    if snidresults is None:
        return None
    
    # snid typing
    if typing is not None:
        snid_typing = SNIDPARSE_TYPING.get(typing,None)
        if snid_typing is None:
            warnings.warn(f"Could not parse input {typing} typing")
        elif verbose:
            print(f"using snid_typing:{snid_typing} as input typing:{typing}")

    else:
        snid_typing = None
        
    if snid_typing in ["*","all", "any"]:
        snid_typing = None

    elif snid_typing in ["snia", "sn ia", "Ia"]:
        snid_typing = snidresults[snidresults["typing"] == "Ia"]["type"].unique()

    elif snid_typing is not None and "!" in snid_typing:
        not_these = snid_typing.split("!")[-1].split(",")
        snid_typing = snidresults[(snidresults["typing"] == "Ia") & 
                             ~(snidresults["type"].isin(not_these))]["type"].unique()
            
    # else typing is None            
    if snid_typing is not None:
        snidresults = snidresults[snidresults["type"].isin(np.atleast_1d(snid_typing))]        

    #
    # select data to parse
    #
    # - grade
    if grade is not None:
        snidresults = snidresults[snidresults["grade"].isin(np.atleast_1d(grade))]
        
    # - rlap        
    if rlap_range is not None:
        # handles None as no limit
        rlap_range  = rlap_range.copy()
        if rlap_range[0] is None: rlap_range[0] = 0
        if rlap_range[1] is None: rlap_range[1] = np.inf
        snidresults = snidresults[snidresults["rlap"].between(*rlap_range)]
    
    # - nfirst
    if n_range is not None:
        n_range = n_range.copy()
        if n_range[0] is None: n_range[0] = 0
        if n_range[1] is None: n_range[1] = +np.inf
        snidresults = snidresults[snidresults["no."].astype(int).between(*n_range)]
        
        
    return snidresults

def get_target_snidredshift(ztfname, 
                                weight_by="rlap",
                                typing=None, redshift_err="nmad",
                                rlap_range=[5, None], n_range=[None, 30],
                                instrument="*", 
                                snid_offset=0.0013,
                                apply_zoffset=True,
                                flag_issues=True,
                            as_dataframe=False,
                                **kwargs):
    """ 
    Returns
    -------
    list
        (redshift, redshift_err)
    
    """
    sres = get_target_snidres(ztfname, typing=typing, 
                                rlap_range=rlap_range, n_range=n_range, 
                                instrument=instrument,
                                **kwargs)
    if sres is None or len(sres)==0:
        return np.NaN, np.NaN
    
    # correct for instrument offset.
    if apply_zoffset:
        sres = sres.merge(INSTRUMENT_ZOFFSET, on="instrument", how="left"
                         ).fillna({"zoffset":0})
        sres["z"] = sres["z"] + sres["zoffset"]
        
    # correct for snid offset
    if snid_offset is not None:
        sres["z"] -= snid_offset
    
    # Special output
    if as_dataframe:
        return sres
    # Special output    
    
    if weight_by is not None:
        weights = sres[weight_by]
        
    if np.sum(weights)==0:
        print(f"{ztfname} weights=0")
        return np.NaN, np.NaN
    

    redshift = np.average(sres["z"], weights=weights)    
    if redshift_err == "nmad":
        if len(sres) < 3:
            dredshift = np.NaN
        else:
            dredshift = median_abs_deviation(sres["z"])
    else:
        dredshift = getattr(np, redshift_err)(sres)
        
    if flag_issues:
        issue_report = check_for_issues(sres)
        if len(issue_report)>0:
            print(f" {ztfname} ".center(30, "-"))
            print( issue_report )
            print(" end report ".center(30, "-"))
            
    return redshift, dredshift
                            
    
def check_for_issues(snidres):
    """ """
    mean_, median_, std_, nmad_ = snidres["z"].agg([np.mean, np.median, np.std, median_abs_deviation])
    nentries = len(snidres)
    
    issues = []
    # too large scatter
    if nentries < 3:
        issues.append("less than 3 entries")
    else: # otherwise nothing make sense
        if std_ > 0.01:
            issues.append("large STD (>0.01)")

        if std_ > 0.05:
            issues.append("WARNING very large STD (>0.05)")

        if std_ > 5*nmad_:
            issues.append(f"STD {std_:.3f}> 5* nMAD {nmad_:.3f}")
    
    return issues






import pandas
from ztfidr import io
import numpy as np


import os


_ORIGINAL_PATH = os.path.abspath( os.path.join(io.IDR_PATH,"../.."))

ZTF_HOST_REPO = os.path.join(_ORIGINAL_PATH, "ztfdr2_host_visualisation/cutouts/")
host_globalfile = os.path.join(io.IDR_PATH, "tables/.dataset_creation/host_prop/host_photometry/ztfdr2_global_imcat_flux.csv")

HOSTDATA  = pandas.read_csv(host_globalfile, index_col=0)


WAVEEFF={'ps1::g': 4866.46,
         'ps1::r': 6214.62,
         'ps1::i': 7544.57,
         'ps1::z': 8679.47,
         'ps1::y': 9633.25,
         'sdss::u': 3594.33,
         'sdss::g': 4717.6,
         'sdss::r': 6186.8,
         'sdss::i': 7506.21,
         'sdss::z': 8918.3,
         '2mass::j': 12410.29,
         '2mass::h': 16513.45,
         '2mass::ks': 21656.09,
         'galex::fuv':1516,
         'galex::nuv':2267,
          }
    
markers= {"ps1":"s",
          "sdss":"o",
          "2mass":"d",
          "galex":"H"}





def jansky_to_abmag(flux, flux_err):
    """
    $m_{\text{AB}}=-2.5\log _{10}f_{\nu }+8.90.$
    """
    
    mag = -2.5*np.log10(flux) + 8.90
    if flux_err is None:
        dmag = np.NaN
    else:
        dmag = +2.5/np.log(10) * np.abs(flux_err / flux)
            
    return mag, dmag
    



def get_target_hostpng(targetname, test_exists=True):
    """ """
    paths = {"cutout1":os.path.join(ZTF_HOST_REPO, f"host_match/{targetname}_zoom1_host_ls.png"),
             "cutout6":os.path.join(ZTF_HOST_REPO, f"host_match/{targetname}_zoom6_host_ls.png"),
             "photimg":os.path.join(ZTF_HOST_REPO, f"host_phot/images/{targetname}/global_PS1_r.jpg")
            }
    if test_exists:
        paths = {k:v if os.path.isfile(v) else None for k,v in paths.items()}
    return paths
    

def get_target_globalphotometry(targetname, add_mag=True):
    """ """    
    datadict= {bandname:[WAVEEFF.get(f"{bandname}", np.NaN), 
                          HOSTDATA.loc[targetname][f"{band}"],
                          HOSTDATA.loc[targetname][f"{band}_err"],
                         ] 
                for bandname, band in zip( ["ps1::g","ps1::r","ps1::i","ps1::z","ps1::y",
                                            "sdss::u","sdss::g","sdss::r","sdss::i","sdss::z",
                                            "2mass::j" ,"2mass::h","2mass::ks",
                                            "galex::fuv", "galex::nuv"],
                                           ["PS1g", "PS1r", "PS1i", "PS1z", "PS1y", 
                                            "SDSSu","SDSSg","SDSSr","SDSSi","SDSSz",
                                            "2MASSJ","2MASSH","2MASSKs",
                                           "GALEXFUV","GALEXNUV"]
                                          )
               }
    data = pandas.DataFrame.from_dict(datadict, orient="index", columns=["waveeff","flux","flux_err"])    
    if add_mag:
        mag, magerr = jansky_to_abmag(data["flux"],data["flux_err"])
        data["mag"] = mag
        data["mag_err"] = magerr
        
    return data


def show_host(targetname, inmag=True, fig=None):
    """ """
    hostimg = get_target_hostpng(targetname)
    data = get_target_globalphotometry(targetname).reset_index(names="filter").copy()
    
    from matplotlib.colors import to_rgba
    from matplotlib.patches import Rectangle
    from matplotlib.image import imread
    if hostimg["cutout1"] is not None:
        img1 = imread(hostimg["cutout1"])
        shape = np.asarray(img1.shape[:2])
        width = shape/6.
    else:
        img1 = None
        shape = None
    if hostimg["cutout6"] is not None:
        img6 = imread(hostimg["cutout6"])
    else:
        img6 = None
    
    if fig is None:
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=[7,2])

    # Image Patches
    ax1 = fig.add_subplot(143)
    if img1 is not None:
        ax1.imshow(img1)
        rect = Rectangle(shape/2.-width/2, width=width[0], height=width[1],
                    facecolor="None", edgecolor="k", lw=0.5)
        ax1.add_patch(rect)
    else:
        ax1.text(0.5,0.5, "no image", transform=ax1.transAxes, va="center", ha="center")

    ax2 = fig.add_subplot(144)
    if img6 is not None:
        ax2.imshow(img6)
    else:
        ax2.text(0.5,0.5, "no image", transform=ax2.transAxes, va="center", ha="center")

    [ax_.set(xticks=[], yticks=[]) for ax_ in fig.axes]

    # Global mag
    ax =  fig.add_subplot(121)

    key = "mag" if inmag else "flux"

    
    data["markers"] = data["filter"].str.split("::", expand=True)[0].apply(markers.get).values

    # remove NaN
    data = data[~data[key].isna()]

    
    for marker, indexes in data.groupby("markers").groups.items():
        ax.scatter(data.loc[indexes]["waveeff"], data.loc[indexes][key], marker=marker, 
                  facecolors=to_rgba("C1", 0.8), edgecolors="0.5", zorder=6)

        ax.errorbar(data.loc[indexes]["waveeff"], data.loc[indexes][key], 
                    yerr=data.loc[indexes][f"{key}_err"],
                    marker="None", ls="None", ecolor="0.5", zorder=5)


    if not inmag:
        ax.set_yticks([0])
        ax.axhline(0, ls="--", color="k", lw=0.5)
        label = "flux [Jy]"
    else:
        ax.invert_yaxis()
        ylim = ax.get_ylim()
        ax.set_ylim( np.min([23, ylim[0]]) ,np.max([ylim[1],13]))
        label = "mag [AB]"

    ax.tick_params(labelsize="small")

    ax.set_xticks(np.linspace(1_000, 23_000, 5))
    ax.set_ylabel(label)
    ax.set_xlabel("wavelength")

    # fancy
    fig.tight_layout()
    return fig




from scipy import stats


noise_model = {
    # SALT paramerers
    "x1": { "func": stats.alpha.rvs,
            "kwargs": {"a":3.46e+00, "loc":-1.35e-01, "scale":9.81e-01}
          },
    "c": { "func": stats.alpha.rvs,
            "kwargs": {"a":3.27e+00, "loc":1.71e-02, "scale":5.03e-02}
          },

    # derived
    "magobs": { "func": stats.alpha.rvs,
            "kwargs": { "a":2.45e+00, "loc":2.54e-02, "scale":3.18e-02}
            },
    # Environments
    "mass": { "func": stats.alpha.rvs,
            "kwargs": { "a":3.58e+00, "loc":1.01e-01, "scale":1.55e-03}
            },
    "localcolor": { "func": stats.alpha.rvs,
                    "kwargs":{"a":3e-7, "loc": 0.01, "scale": 0.017}
                }
    }
    
                



import numpy as np

import pandas
from scipy import stats
from scipy.special import legendre


__all__ = ["fit_spectrum"]

def fit_spectrum(spectrum, redshift, figure=True, dirout=None, lbda_window=None, **kwargs):
    """ loads the intance and fit a spectrum 
        
    Parameters
    ----------
    spectrum: ztfidr.Spectrum or str
        spectrum or spectrum filepath to fit

    redshift: float
        initial redshift guess

    figure: bool
        should a figure be created. 
        If False, returned figure is None.

    dirout: str
        path for the directory where results will be stored. 
        if figure is True, the figure will also be stored.
        
    **kwargs fit() options
        
    Returns
    -------
    pandas.Series, pandas.DataFrame
        metadata table and results (value, error, cov_)
    """
    return LineFitter.fit_spectrum(spectrum, redshift, figure=figure, dirout=dirout,
                                       lbda_window=lbda_window, **kwargs)

class LineFitter( object ):
    LINES = {"HA":6562.8,
             "NII":6583.4,
            }
    JOINED_LINES = {"NII":{"lbda":6548.0, "amplitude":1/2.8}
                    }
    
    BACKGROUND_DEGREE = 3 # only affects guess, rest is self consistent.
    LBDA_WINDOW = 150  # value used it not provided.
    
    def __init__(self, spectrum=None, redshift_init=0, lbda_window=None):
        """ """
        self.spectrum = spectrum
        self.redshift_init = redshift_init
        
        if lbda_window is None:
            lbda_window = self.LBDA_WINDOW
        
        self.lbda_window = lbda_window

    @classmethod
    def fit_spectrum(cls, spectrum, redshift, figure=True, dirout=None,
                         lbda_window=None, **kwargs):
        """ loads the intance and fit a spectrum 
        
        Parameters
        ----------
        spectrum: ztfidr.Spectrum or str
            spectrum or spectrum filepath to fit

        redshift: float
            initial redshift guess
            
        figure: bool
            should a figure be created. 
            If False, returned figure is None.

        **kwargs fit() options
        
        Returns
        -------
        (pandas.Series, pandas.DataFrame), figure
            metadata table and results (value, error, cov_)
        """
        if type(spectrum) is str:
            from .spectroscopy import Spectrum
            spectrum = Spectrum.from_filename(spectrum)
            
        # Loads
        this = cls(spectrum, redshift, lbda_window=lbda_window)
        # Fit        
        params = this.fit(**kwargs)
        # Plot
        if figure:
            fig = this.show_data(show_guess=True, 
                                    model_parameters=params["value"])
        else:
            fig = None
        # Store
        if dirout is not None:
            _ = this.store(dirout, fig=fig)
            
        return *this.to_pandas(), fig

    # ---------- #
    # High Level #
    # ---------- #
    def fit(self, guess={}, limits={}, fixed={},
                limit_sigma=[.01, 5], limit_reshift=[0,0.4],
                sigma_pixel=True,
                **kwargs):
        """ fit the model. 

        The fitted data are these from self.fitted_data that is created
        automatically up on first call using get_fitted_data(). 
        
        Parameters
        ----------
        guess: dict
            dictionary to overwrite any default guess parameters
            obtained using self.get_guess_parameters(). 

        limit_sigma: (float, float)
            boundaries of the sigma parameters (emission line 'scale' in lbda-pixel)

        limit_redshift: (float, float)
            boundaries for the fitted redshift.
    
        **kwargs goes to Minuit()

        Returns
        -------
        DataFrame

        See also
        --------
        get_fitted_data: get the data as fitted (truncated as needed and normalize)
        to_pandas: get the fit result in pandas format (meta, result)
        """
        if fixed is None:
            fixed = {}
            
        if limits is None:
            limits = {}
        
        # red input
        if limit_sigma is not None and sigma_pixel:
            average_step = self.fitted_data["lbda"].diff().mean()
            limit_sigma = list(np.asarray(limit_sigma) * average_step)
            sigma = np.sqrt(2**2 + average_step**2) # 2 A corresponds to ~90 km/s gas dispersion
        else:
            sigma = None

        
        from iminuit import Minuit
        from iminuit.util import make_func_code
        # Minuit naming
        setattr(self.__class__.get_logprob, "func_code", make_func_code(self.param_names))

        # create the guesses
        guess_list = self.get_guess_parameters(sigma=sigma)
        guess = {**dict( zip(self.param_names, guess_list) ), **guess}
        guess = {**guess, **fixed} # force the fixed parameters
        self._fit_guess = guess

        #
        # load Minuit
        #
        # -> function to optimize
        self._minuit = Minuit( self.get_logprob, **guess, **kwargs)
        
        # -> parameter limits
        self._minuit.limits["sigma"] = limit_sigma
        self._minuit.limits["redshift"] = limit_reshift
        for k, v in limits.items():
            self._minuit[k] = v

        # -> parameter to fix
        for k, v in fixed.items(): # value set in the guess.
            self._minuit.fixed[k] = True

        # 
        # Run the fit
        #
        self._minuit.migrad()
        self._minuit.hesse()
        
        # output 
        return self.fitted_parameters

    def store(self, dirout, fig=None):
        """ """
        import os
        filename = os.path.basename(self.spectrum.filename).replace('.ascii', f'_linfit.parquet')
        filepath = os.path.join(dirout, filename)
        meta, results = self.to_pandas()
        results.to_parquet(filepath)
        if fig is not None:
            fig.savefig( filepath.replace(".parquet",".pdf") )

        return filepath
    
    def to_pandas(self, restore_norm=True):
        """ convert the minuit output into pandas format

        Parameters
        ----------
        restore_norm: bool
            the fitted data are normalized. Should the returned amplitude parameters
            be restoted into input data units ? 
            If False, returned results will be in fitted_data units ; 
            as for self.fitted_parameters.

        Returns
        -------
        pandas.Series, pandas.DataFrame
            metadata table and results (value, error, cov_)

        See also
        --------
        fitted_parameters: fit results in pandas dataframe
        fit: main fit method.
        """
        results = self.fitted_parameters # dataframe of value, error, cov
        results = results.join( pandas.Series(self._fit_guess, name="guess"))

        meta = pandas.Series({k: getattr(self._minuit,k) for k in ["fval","valid","accurate","nfit","nfcn"]})
        meta["ha_detection"] = (results["value"]/results["error"]).loc["ampl_ha"]
        meta["dof"] = len(self.fitted_data) - len(self.free_parameters)
        meta["targetname"] = self.spectrum.targetname
        
        if self._fitdata_norm != 1 and restore_norm: # that would also work, but useless
            results.loc[results.index.str.startswith("ampl_")] *= self._fitdata_norm
            results.loc[:,results.columns.str.startswith("cov_ampl_")] *=  self._fitdata_norm


        return meta, results

    # --------- #
    #  Method   #
    # --------- #
    def get_fitted_data(self, redshift_init=None,  
                        check_variance=True, var_window_percent=10,
                        normalised=True):
        """ get a subpart of the data where to perform the fit. 
        They are truncated in wavelength to the relevant lbda range and normalized. 

        Parameters
        ----------
        redshift_init: float
            initial redshift guess used to truncate the wavelength range of interest
            (see self.lbda_window parameters).
            If None self.redshift_init is used (recommended)

        check_variance: bool
            should this method fix any potential variance issues. 
            Under or over estimated variance affects the fitted parameter 
            error estimation. Scatter of the data at the fitted_data edges 
            are used for that. (see var_window_percent).
            This also created a variance estimation if no variance exist.

        var_window_percent: float
            = ignored if check_variance is False = 
            the pixel-window (in percent of total fitted_data lbda) used 
            to estimate the expected signal variance ; half from each edges 
            of the fitted_data wavelength range. 
            For instance, var_window_percent=10 means that 10% of the 
            spectrum will be used to estimate the variance (flux std) 
            using the average of the first and last 5%.
            
        normalised: bool
            should the data be normalized (flux.mean) for the fit ?
            (highly recommended)
        
        Returns
        -------
        pandas.DataFrame
            lbda, flux[, variance]
           
        """
        if redshift_init is None:
            redshift_init = self.redshift_init
        
        lines = np.asarray(self.fitted_lines)
        lbda_flag = (self.spectrum.lbda > (lines.min()*(1+redshift_init) - self.lbda_window)) & \
                    (self.spectrum.lbda < (lines.max()*(1+redshift_init) + self.lbda_window))
        
        data = self.spectrum.data[lbda_flag].copy()
        data["lbda_x"] = ((data["lbda"].values - data["lbda"].min())/(data["lbda"].max() - data["lbda"].min()) - 0.5)*2
        if check_variance:
            var_window = int(data.size*var_window_percent/100*0.5)
            var_ = np.mean([data["flux"][:var_window].values.std()**2,
                           data["flux"][-var_window:].values.std()**2])
            
            if  "variance" not in data:
                data["variance"] = var_
                
            else:
                var_in = np.mean([data["variance"][:var_window].values.mean(),
                                  data["variance"][-var_window:].values.mean()])
                data["variance"] *= (var_/var_in)

        if normalised:
            norm = np.nanmean(data["flux"])
            data["flux"] /= norm
            if "variance" in data:
                data["variance"] /= norm**2
            self._fitdata_norm = norm
        else:
            self._fitdata_norm = 1

        return data
        
    # --------- #
    #   Model   #
    # --------- #
    def get_logprob(self, *parameters):
        """ """
        chi2 = self.get_chi2(parameters)
        priors = 0 # to be addid if needed.
        return chi2 + priors
        
    def get_chi2(self, parameters, leastsq=False, data=None):
        """ """
        data = self.get_model(parameters, data=data)
        to_minimize = data["residual"].values**2
        if not leastsq and "variance" in data:
            to_minimize /= data["variance"].values
            
        return np.nansum(to_minimize)
        
    def parse_parameters(self, parameters):
        """ """
        amplitudes = parameters[:self.nlines]
        redshift, sigma_lines, *b_parameters = parameters[self.nlines:]
        return list(amplitudes), redshift, sigma_lines, list(b_parameters)

    
    def get_linemodel(self, lbda, amplitudes, redshift, sigma):
        """ """
        all_amplitudes = np.asarray(self.get_lineamplitudes(amplitudes))
        all_lbda = np.asarray(self.fitted_lines) * (1 + redshift)

        all_model = np.dot(all_amplitudes,
                           stats.norm.pdf(np.asarray(lbda)[:,None], 
                                          loc=all_lbda, 
                                          scale=sigma).T)
        return all_model
    
    def get_lineamplitudes(self, amplitudes):
        """ build full amplitudes, adding given amplitudes to joined ones. """
        dict_ampl = dict(zip(self.fitted_line_names, amplitudes))
        joined_ampl = [self.JOINED_LINES.get(k).get("amplitude", np.NaN)*dict_ampl[k] 
                      for k in self.fitted_line_names if k in self.JOINED_LINES]
        return amplitudes + joined_ampl
    
    @staticmethod
    def get_backgroundmodel(parameters, lbda_x):
        """ """
        background = np.dot(parameters, np.stack([legendre(id_)(lbda_x) 
                                               for id_ in range( len(parameters) )])
                           )
        return background
    
    def get_model(self, parameters, data=None, verbose=False, **kwargs):
        """ """
        if data is None:
            data = self.fitted_data
            
        # Parse inputs
        amplitudes, redshift, sigma_lines, b_parameters = self.parse_parameters(parameters)
        if verbose:
            print(f"amplitudes: {amplitudes}")
            print(f"redshift: {redshift}")
            print(f"sigma_lines: {sigma_lines}")
            print(f"b_parameters: {b_parameters}")            
        
        
        # background
        background = self.get_backgroundmodel(b_parameters, data["lbda_x"])
        # lines
        lines = self.get_linemodel(data["lbda"], amplitudes, redshift, sigma_lines)
        # add it to the dataframe
        data["model"] = background + lines
        data["residual"] = data["model"] - data["flux"]
        
        return data
    
    def get_guess_parameters(self, sigma=None, redshift_buffer=10, **kwargs):
        """ 

        """
        #
        # background
        #
        b_params = np.zeros(self.BACKGROUND_DEGREE)
        b_params[0] = self.fitted_data["flux"].median() # constant. 
        if self.BACKGROUND_DEGREE>1:
            # devided by 2 because lbda_x defined between -1 and 1
            b_params[1] = (self.fitted_data["flux"].iloc[-5:].mean() - self.fitted_data["flux"].iloc[:5].mean())/2
            
        #
        # redshift
        #
        top_line_args = self.fitted_data["flux"].argmax()
                
        if redshift_buffer is not None:
            ha_arg = np.argmin(np.abs(self.fitted_data["lbda"] - self.LINES["HA"]*(1+self.redshift_init)))
            # 4 pixels buffer
            buffer_size = 2
            local_argmax = self.fitted_data.iloc[ha_arg-buffer_size : ha_arg+buffer_size]["flux"].argmax()
            top_line_args = ha_arg - buffer_size + local_argmax

        redshift = (self.fitted_data.iloc[top_line_args]["lbda"] / self.LINES["HA"]) - 1

        #
        # sigma
        #
        if sigma is None:
            average_step = self.fitted_data["lbda"].diff().mean()
            sigma = np.sqrt(2**2 + average_step**2) # 2 A corresponds to ~90 km/s gas dispersion
        # - Nothing do do
        
        #
        # Amplitudes
        # 
        ha_amplitude = (self.fitted_data.iloc[top_line_args]["flux"] - b_params[0]) * np.sqrt(2*np.pi)*sigma
        amplitudes = np.ones(self.nlines)*ha_amplitude 
        amplitudes[1:] /= 3 # all lines at a third of Ha
        # returns
        return *amplitudes, redshift, sigma, *b_params
        
        
    def show_data(self, ax=None, show_lines=True, show_error=True, 
                 show_guess=False, model_parameters=None, 
                  legend=True):
        """ """
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=[7,5])
        ax = fig.add_subplot(111)
        
        data = self.get_fitted_data().astype(float) # does a copy
        ax.plot(data.lbda, data.flux, label="data")
        ax.set_xlabel("wavelength [A]", fontsize="large")
        ax.set_ylabel("flux []", fontsize="large")
        
        
        if "variance" in data and show_error:
            err = np.sqrt(data["variance"])
            ax.fill_between(data.lbda, data.flux+err, data.flux-err, alpha=0.1)
        
        if show_lines:
            redshift = self.redshift_init
            _ = [ax.axvline(line*(1+redshift), ls="--", color="0.7")
                 for line in self.fitted_lines]
            
        if show_guess:
            if hasattr(self, "_fit_guess"):
                parameters = [self._fit_guess[k] for k in self.param_names]
            else:
                parameters = self.get_guess_parameters()
                
            data = self.get_model(parameters, data=data)
            ax.plot(data["lbda"], data["model"], ls=":", label="guess")
            
        if model_parameters is not None:
            data = self.get_model(model_parameters, data=data)
            ax.plot(data["lbda"], data["model"], ls="-", label="model")
            
        if legend:
            ax.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left',
                      ncol=3, mode="expand", borderaxespad=0.)
        return fig
        
    # ============ #
    #  Properties  #
    # ============ #
    @property
    def nlines(self):
        """ """
        return len(self.LINES)
    
    @property
    def fitted_line_names(self):
        """ """
        return list(self.LINES.keys())

    @property
    def fitted_data(self):
        """ """
        if not hasattr(self, "_fitted_data") or self._fitted_data is None:
            self._fitted_data = self.get_fitted_data()
            
        return self._fitted_data
    
    @property
    def fitted_lines(self):
        """ """
        base_lines = [self.LINES.get(k, np.NaN) for k in self.fitted_line_names]
        joined_lines = [self.JOINED_LINES.get(k).get("lbda", np.NaN) 
                        for k in self.fitted_line_names if k in self.JOINED_LINES]
        return base_lines+joined_lines
    
    @property
    def fitted_parameters(self):
        """ """
        dataout = pandas.DataFrame(self._minuit.values, columns=["value"], index=self._minuit.parameters)
        dataout["error"] = np.sqrt(np.diag(self._minuit.covariance))
        cov = pandas.DataFrame( self._minuit.covariance, 
                               index=dataout.index, columns="cov_"+dataout.index)
        
        results = dataout.join(cov)
        return results

    @property
    def free_parameters(self):
        """ """
        if not hasattr(self,"_minuit"):
            raise AttributeError("Minuit has not run yet.")
        
        return [k for k,v in self._minuit.fixed.to_dict().items() if not v]
    
    @property
    def param_names(self):
        """ """
        return [f"ampl_{k.lower()}" for k in self.fitted_line_names] + \
               ["redshift", "sigma"] + \
               [f"ampl_back{k}" for k in range(self.BACKGROUND_DEGREE)]
    
